/**
 * Copyright (C) 2005-2006 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.email.admin;

import java.awt.Component;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.Serializable;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.JCheckBox;
import javax.swing.SwingConstants;
import javax.swing.border.TitledBorder;

import org.apache.commons.lang.StringUtils;

import com.funambol.admin.AdminException;
import com.funambol.admin.ui.ConnectorManagementPanel;


/**
 * This class implements the connector configuration panel
 * for a Email Connector component
 *
 *
 * @version $Id: EmailConnectorConfigPanel.java,v 1.2 2006/02/06 08:58:40 gbmiglia Exp $
 */
public class EmailConnectorConfigPanel
               extends ConnectorManagementPanel
               implements Serializable {

    // --------------------------------------------------------------- Constants

    public static final String NAME_ALLOWED_CHARS =
            "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890-_.";

    // ------------------------------------------------------------ Private data

    private JPanel serverPanel;
    private JPanel funambolPanel;
    private JPanel propertyPanel;

    private JLabel     inHostLabel;
    private JTextField inHostValue;
    private JLabel     outHostLabel;
    private JTextField outHostValue;
    private JLabel     protocolLabel;
    private JTextField protocolValue;
    private JLabel     portLabel;
    private JTextField portValue;

    private JLabel     dataSourceLabel;
    private JTextField dataSourceValue;

    private JLabel     filterLabel;
    private JCheckBox  filterValue;

    private JButton confirmButton = new JButton();

    // ------------------------------------------------------------ Constructors


    public EmailConnectorConfigPanel() {
      init();
    }


    // --------------------------------------------------------- Private methods

    /**
     * Create the panel
     * @throws Exception if error occures during creation of the panel
     */
    private void init() {

        JLabel title = new JLabel();

        serverPanel = new JPanel();

        inHostLabel   = new JLabel();
        inHostValue   = new JTextField();
        outHostLabel  = new JLabel();
        outHostValue  = new JTextField();
        protocolLabel = new JLabel();
        protocolValue = new JTextField();
        portLabel     = new JLabel();
        portValue     = new JTextField();

        funambolPanel = new JPanel();

        dataSourceLabel = new JLabel();
        dataSourceValue = new JTextField();

        propertyPanel = new JPanel();

        filterLabel   = new JLabel();
        filterValue   = new JCheckBox();


        setLayout(null);

        title.setFont(titlePanelFont);
        title.setText("Funambol Email Connector");
        title.setBounds(new Rectangle(14, 5, 316, 28));
        title.setAlignmentX(SwingConstants.CENTER);
        title.setBorder(new TitledBorder(""));

        serverPanel.setLayout(null);
        serverPanel.setBorder(new TitledBorder("BackEnd Server Configuration"));

        inHostLabel.setText("Incoming Server:");
        serverPanel.add(inHostLabel);
        inHostLabel.setBounds(10, 20, 116, 15);
        serverPanel.add(inHostValue);
        inHostValue.setBounds(150, 20, 220, 19);

        outHostLabel.setText("Outgoing Server:");
        serverPanel.add(outHostLabel);
        outHostLabel.setBounds(10, 50, 131, 15);
        serverPanel.add(outHostValue);
        outHostValue.setBounds(150, 50, 220, 19);

        protocolLabel.setText("Protocol (Officer):");
        serverPanel.add(protocolLabel);
        protocolLabel.setBounds(10, 80, 131, 15);
        serverPanel.add(protocolValue);
        protocolValue.setBounds(150, 80, 220, 19);

        portLabel.setText("Port (Officer):");
        serverPanel.add(portLabel);
        portLabel.setBounds(10, 110, 131, 15);
        serverPanel.add(portValue);
        portValue.setBounds(150, 110, 220, 19);

        add(serverPanel);
        serverPanel.setBounds(10, 50, 380, 140);


        funambolPanel.setLayout(null);
        funambolPanel.setBorder(new TitledBorder("Funambol Configuration"));

        dataSourceLabel.setText("Funambol DataSource:");
        funambolPanel.add(dataSourceLabel);
        dataSourceLabel.setBounds(10, 20, 131, 15);
        funambolPanel.add(dataSourceValue);
        dataSourceValue.setBounds(150, 20, 220, 19);

        add(funambolPanel);
        funambolPanel.setBounds(10, 200, 380, 50);


        propertyPanel.setLayout(null);
        propertyPanel.setBorder(new TitledBorder("Property Configuration"));

        filterLabel.setText("Apply Filter: ");
        propertyPanel.add(filterLabel);
        filterLabel.setBounds(new Rectangle(10, 20, 150, 19));
        filterValue.setSelected(false);
        propertyPanel.add(filterValue);
        filterValue.setBounds(150, 20, 220, 19);

        add(propertyPanel);
        propertyPanel.setBounds(10, 260, 380, 50);



        confirmButton.setFont(defaultFont);
        confirmButton.setText("Save");
        confirmButton.setBounds(160, 400, 70, 25);

        confirmButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event ) {
                try {
                    validateValues();
                    getValues();
                    EmailConnectorConfigPanel.this.actionPerformed(
                             new ActionEvent(EmailConnectorConfigPanel.this,
                                             ACTION_EVENT_UPDATE,
                                             event.getActionCommand()));
                } catch (Exception e) {
                    notifyError(new AdminException(e.getMessage()));
                }
            }
        });

        add(confirmButton, null);

        //
        // Setting font...
        //
        Component[] components = getComponents();
        for (int i=0; (components != null) && (i<components.length); ++i) {
            components[i].setFont(defaultFont);
        }

        //
        // We add it as the last one so that the font won't be changed
        //
        add(title);
  }

  /**
   *
   */
  public void updateForm() {

      EmailConnectorConfig emailConf = (EmailConnectorConfig)getConfiguration();

      inHostValue.setText(emailConf.getInHost());
      outHostValue.setText(emailConf.getOutHost());

      protocolValue.setText(emailConf.getProtocol());
      portValue.setText(""+emailConf.getPort());

      dataSourceValue.setText(emailConf.getDataSource());

      if (emailConf.getFilter()) {
          filterValue.setSelected (true) ;
      } else {
          filterValue.setSelected (false) ;
      }

  }

  // ----------------------------------------------------------- Private methods

  /**
   * Checks if the values provided by the user are all valid. In caso of errors,
   * a IllegalArgumentException is thrown.
   *
   * @throws IllegalArgumentException if:
   *         <ul>
   *         <li>anyvalue is not in the form <host>:<port>
   *         </ul>
   */
  private void validateValues() throws IllegalArgumentException {
      String value = null;

      // in host
      value = inHostValue.getText().trim();
      if (StringUtils.isEmpty(value)) {
          throw new IllegalArgumentException("Field 'Incoming Host' cannot be empty");
      }

      // out host
      value = outHostValue.getText().trim();
      if (StringUtils.isEmpty(value)) {
          throw new IllegalArgumentException("Field 'Outgoing Host' cannot be empty");
      }

      // protocol
      value = protocolValue.getText().trim();
      if (StringUtils.isEmpty(value)) {
          throw new IllegalArgumentException("Field 'Protocol' cannot be empty");
      }

      // port
      value = portValue.getText().trim();
      if (StringUtils.isEmpty(value)) {
          throw new IllegalArgumentException("Field 'Port' cannot be empty");
      }
      try {
          Integer.parseInt(value);
      } catch (NumberFormatException e) {
          throw new IllegalArgumentException("Field 'Port' must be numeric");
      }


      // datasource
      value = dataSourceValue.getText().trim();
      if (StringUtils.isEmpty(value)) {
          throw new IllegalArgumentException("Field 'Sync4j DataSource' cannot be empty");
      }


  }



  /**
   * Set properties with the values provided by the user.
   */
  private void getValues() {

      EmailConnectorConfig emailConf = (EmailConnectorConfig)getConfiguration();

      emailConf.setInHost(inHostValue.getText().trim());
      emailConf.setOutHost(outHostValue.getText().trim());

      emailConf.setProtocol(protocolValue.getText().trim());
      emailConf.setPort(Integer.parseInt(portValue.getText().trim()));

      emailConf.setDataSource(dataSourceValue.getText().trim());

      if (filterValue.isSelected()) {
         emailConf.setFilter(true) ;
      } else {
         emailConf.setFilter(false) ;
      }


  }

}
