/**
 * Copyright (C) 2005-2006 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.email.engine.source;

import java.io.Serializable;
import java.io.ByteArrayInputStream;
import java.util.List;
import java.util.TimeZone;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.Date;
import java.util.StringTokenizer;
import java.net.UnknownHostException;
import java.net.InetAddress;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.security.Principal;
import java.security.GeneralSecurityException;
import java.security.Signature;

import com.funambol.framework.engine.SyncItem;
import com.funambol.framework.engine.SyncItemKey;
import com.funambol.framework.engine.source.SyncSource;
import com.funambol.framework.engine.source.SyncSourceException;
import com.funambol.framework.security.Sync4jPrincipal;
import com.funambol.framework.server.Sync4jDevice;
import com.funambol.framework.server.store.PersistentStoreException;
import com.funambol.framework.server.store.PersistentStore;
import com.funambol.framework.engine.SyncItemImpl;
import com.funambol.framework.engine.SyncProperty;
import com.funambol.framework.filter.FilterClause;
import com.funambol.framework.filter.LogicalClause;
import com.funambol.framework.filter.WhereClause;
import com.funambol.framework.filter.FieldClause;
import com.funambol.framework.filter.IDClause;
import com.funambol.framework.filter.Clause;
import com.funambol.framework.tools.Base64;
import com.funambol.framework.core.Property;
import com.funambol.server.config.Configuration;
import com.funambol.foundation.pdi.utils.TimeUtils;

import com.funambol.email.pdi.parser.XMLMailParser;
import com.funambol.email.pdi.converter.MailToXML;
import com.funambol.email.pdi.mail.*;
import com.funambol.email.cache.CrcCachedSyncSource;
import com.funambol.email.exception.EmailAccessException;
import com.funambol.email.exception.EntityException;
import com.funambol.email.exception.EmailConfigException;
import com.funambol.email.util.Def;
import com.funambol.email.util.Utility;
import com.funambol.email.transport.IMailServerWrapper;
import com.funambol.email.transport.MailServerWrapperFactory;
import com.funambol.email.items.model.SendResult;
import com.funambol.email.items.manager.EntityManager;
import com.funambol.email.admin.EmailConnectorConfig;


/**
 *
 * This class implements common methods of
 * <pre>
 *  EmailIMAPSyncSource
 *  EmailPOPSyncSource
 * </pre>
 *
 *
 * version $Id: EmailSyncSource.java,v 1.2 2006/02/06 09:00:20 gbmiglia Exp $
 *
 */
public abstract class EmailSyncSource
                extends CrcCachedSyncSource
                implements SyncSource, Serializable {

    // --------------------------------------------------------------- Constants
/*
    protected final String LICENSE_FILE_NAME =
            "sync4j-email-connector.license";

    protected final String LICENSE_ERROR_MESSAGE =
            "Please contact sales@funambol.com.";
*/

    protected Logger log = Logger.getLogger(Def.LOGGER_NAME);

    // -------------------------------------------------------------- Properties

    /**
     * Property for sync session
     */
    protected Sync4jDevice device  = null;

    /**
     * Property for sync session
     */
    protected String  deviceTimeZoneDescr = null ;

    /**
     * Property for sync session
     */
    protected TimeZone deviceTimeZone  = null;

    /**
     * Property for sync session
     */
    protected String deviceCharset  = null  ;

    /**
     * Property Principal
     */
    protected Principal principal = null;

    /**
     * Property syncMode
     */
    protected int syncMode ;

    /**
     * wrapper for the Mail Server
     */
    protected IMailServerWrapper mswf = null;

    /**
     *
     */
    protected EntityManager em = null;


    /**
     * Property Incoming Server
     * get from the Connector Panel
     */
    protected String inHost = "";

    /**
     * Property Outgoing Server
     * get from the Connector Panel
     */
    protected String outHost = "";


    /**
     * filter
     */
    protected  int    folderFilter = 0;   // def. all folders
    protected  String idFilter     = null; // def. no id filter
    protected  Date   timeFilter   = null; // def. no time filter
    protected  int    sizeFilter   = 21;   // def. get all

    protected  int    numBytes     = 0;

    protected  boolean isHeaderFieldSet = false;
    protected  boolean isBodyFieldSet   = false;
    protected  boolean isAttachFieldSet = false;




    /**
     * Should the channel be ssl?
     */
    protected boolean ssl = false;

    /**
     * Getter for property ssl.
     * @return Value of property ssl.
     */
    public boolean isSSL() {
        return ssl;
    }

    /**
     * Setter for property ssl.
     * @param ssl New value of property ssl.
     */
    public void setSSL(boolean ssl) {
        this.ssl = ssl;
    }

    /**
     * Property Incoming Server Port
     */
    protected String inPort = "";

    /**
     * Getter for property Incoming Mail Server Port.
     * @return Value of property Incoming Mail Server Port.
     */
    public String getInPort() {
        return inPort;
    }

    /**
     * Setter for property Incoming Mail Server Port.
     * @param inPort String New value of property Incoming Mail Server Port.
     */
    public void setInPort(String inPort) {
        this.inPort = inPort;
    }



    /**
     * Property Mail Server Server protocol
     */
    protected String protocol = "";

    /**
     * Getter for property Mail Server protocol.
     * @return Value of property Mail Server protocol.
     */
    public String getProtocol() {
        return this.protocol;
    }

    /**
     * Setter for property Mail Server protocol.
     * @param protocol String. New value of property Mail Server protocol.
     */
    public void setProtocol(String protocol) {
        this.protocol = protocol;
    }

    /**
     * Property Mail Server Language
     */
    protected String language = Def.LANG_EN;

    /**
     * Getter for property Mail Server Language
     * @return Value of property Mail Server Language
     */
    public String getLanguage() {
        return language;
    }

    /**
     * Setter for property Mail Server Language
     *
     * @param language String New value of property Mail Server Language
     */
    public void setLanguage(String language) {
        this.language = language;
    }

    /**
     * Should the content be encoded?
     */
    protected boolean encode = false;

    /**
     * Getter for property encode.
     * @return Value of property encode.
     */
    public boolean isEncode() {
        return encode;
    }

    /**
     * Setter for property encode.
     * @param encode New value of property encode.
     */
    public void setEncode(boolean encode) {
        this.encode = encode;
    }


  // ------------------------------------------------------------ Public Methods

  /**
   * @see SyncSource
   * set the filter
   *
   * folder filter
   *   T D S O I
   *   0 0 0 0 1 -->  1 inbox
   *   0 0 0 1 1 -->  3 inbox + outbox
   *   0 0 1 1 1 -->  7 inbox + outbox + sent
   *   ...
   *   1 1 1 1 1 --> 31 all
   *
   * time filter (timeFilter)
   *   null              --> all
   *   "i.e. 20-10-2005" --> filter
   *
   * size filter (sizeFilter / numbytes)
   *
   *   000001 = 1   FILTER_SIZE_H
   *   000011 = 3   FILTER_SIZE_HPERC
   *
   *   000101 = 5   FILTER_SIZE_H_B
   *   001111 = 15  FILTER_SIZE_H_BPERC
   *
   *   010101 = 21  FILTER_SIZE_H_B_A
   *   111111 = 63  FILTER_SIZE_H_B_APERC
   *
   *
   * @param principal Principal
   * @param syncMode int
   * @param filterClause FilterClause
   * @throws SyncSourceException
   */
  public void beginSync(Principal principal,
                        int syncMode,
                        FilterClause filterClause)
      throws SyncSourceException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("EmailSyncSource - Start beginSync [" + sourceURI + "]");
        }

        // get Connector parameters
        boolean filterActivation = false;
        try {
          inHost  = EmailConnectorConfig.getConfigInstance().getInHost();
          outHost = EmailConnectorConfig.getConfigInstance().getOutHost();
          filterActivation  = EmailConnectorConfig.getConfigInstance().getFilter();
        } catch (EmailConfigException e) {
            throw new SyncSourceException("Error Getting Connector Parameters:" +
                                          e.getMessage(), e);
        }

        // set filter
        if (filterActivation && (filterClause != null)) {
           if (log.isLoggable(Level.FINEST)) {
               log.finest("GET Filter clause!!");
           }
           // first cluse is Logical
           LogicalClause logicalClause =  filterClause.getClause();
           setFilter(logicalClause);
        } else {
           if (log.isLoggable(Level.FINEST)) {
               log.finest("NO Filter clause!!");
           }
        }

        // set common properties
        this.principal = principal;
        this.syncMode  = syncMode;

        String principalId = null;
        String username    = null;
        String credentials = null;
        String password    = null;

        principalId = ((Sync4jPrincipal) this.principal).getId();
        username    = ((Sync4jPrincipal) this.principal).getUsername();
        credentials = ((Sync4jPrincipal) this.principal).getEncodedCredentials();
        password    = Utility.getPassword(username, credentials);

        if (log.isLoggable(Level.FINEST)) {
            log.finest("\nCOORDS " +
                       "\nprincipalId  " + principalId +
                       "\ninHost       " + this.inHost +
                       "\noutHost      " + this.outHost +
                       "\nprotocol     " + this.protocol +
                       "\nlanguage     " + this.language +
                       "\nfilter       " + filterActivation +
                       "\nlogin        " + username +
                       "\nFILTER " +
                       "\nfolderFilter " + this.folderFilter +
                       "\nidFilter     " + this.idFilter +
                       "\ntimeFilter   " + this.timeFilter +
                       "\nsizeFilter   " + this.sizeFilter +
                       "\nnumBytes     " + this.numBytes  );
        }


        // open Mail Server connection
        this.mswf = MailServerWrapperFactory.getMailServerWrapper(this.protocol);

        try {
            this.mswf.openConnection(outHost,
                                     inHost,
                                     inPort,
                                     language,
                                     username,
                                     password,
                                     false);

        } catch (EmailAccessException e) {
            throw new SyncSourceException("Error Connecting Mail Server:" +
                                          e.getMessage(), e);
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("call Cached SyncSource");
        }

        super.beginSync(this.principal, syncMode, filterClause);

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Set Device Parameters");
        }

        String deviceId = null;
        try {
            deviceId = ((Sync4jPrincipal) this.principal).getDeviceId();
            device = getDevice(deviceId);

            String timezone      = device.getTimeZone();
            if (device.getConvertDate ()) {
                if (timezone != null && timezone.length() > 0) {
                    deviceTimeZoneDescr = timezone;
                    deviceTimeZone = TimeZone.getTimeZone(deviceTimeZoneDescr);
                }
            }
            deviceCharset        = device.getCharset();

        } catch (PersistentStoreException ex) {
            throw new SyncSourceException("Error reading device '" + deviceId + "'", ex);
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("EmailSyncSource - End beginSync [" + sourceURI + "]");
        }

  }

    /**
     *
     * @param syncItemKey SyncItemKey
     * @return char
     * @throws SyncSourceException
     */
     /*
    public char getSyncItemStateFromId(SyncItemKey syncItemKey)
       throws SyncSourceException {

         char status = super.getSyncItemStateFromId(syncItemKey);
         return status;
    }
    */


  /**
   * @see SyncSource
   *
   * @throws SyncSourceException
   */
  public void endSync() throws SyncSourceException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("EmailSyncSource - start EndSync [" + sourceURI + "]");
        }

        // reset the cache
        super.endSync();


        // delete from OUTBOX the sent mail
        int sentLen    = itemsSent.size();
        if (log.isLoggable(Level.FINEST)) {
            log.finest("EmailSyncSource - sent mail to move: " + sentLen );
        }
        SendResult sr  = null;
        String  GUID   = null;
        String  LUID   = null;
        boolean isSent = false;
        boolean isPOP  = false;
        String FMID    = null;

        for (int i=0; i<sentLen; i++){

            sr = (SendResult)itemsSent.get(i);

            GUID   = sr.getGUID();
            LUID   = sr.getLUID();
            isSent = sr.isSent();
            isPOP  = sr.isPOP();

            if (log.isLoggable(Level.FINEST)) {
               log.finest(" sent mail guid:" + GUID + " luid: " + LUID +
                          " sent: " + isSent + " isPop: " + isPOP);
            }

            if (isPOP){ // POP3 protocol
               //
               // at the end of the first sync the connector inserts a row in the
               // cache table.
               //
               // in the second sync the cache system find an id in the cache
               // for the outbox document but nothing in the server oubox folder
               // Then the server sends a DELETE command for the oubox folder
               //
               if (isSent && GUID != null){
                   try {
                     super.insertSentItem(GUID);
                   } catch (Exception e) {
                     throw new SyncSourceException("Error handling sent item ", e);
                   }
               }
            } else { // IMAP protocol
               //
               // at the end of the first sync the connector moves message from
               // outbox to sent folder
               //
               // in the second sync the cache system finds an id in the cache
               // for the outbox document but nothing in the server outbox folder
               // then the server sends a DELETE command for the oubox folder
               //
               // morover the cache system finds a document in the server ent folder
               // but nothing in the cache.
               // Then the server sends a ADD command for the sent folder
               //
               if (isSent && GUID != null){
                   FMID = Utility.getFMID(GUID);
                   try {
                     this.em = this.getEntityManager();
                     this.em.moveSentItem(FMID, this.mswf);
                   } catch (Exception e) {
                     throw new SyncSourceException("Error handling sent item ", e);
                   }
               }
            }

        }


        // close Mail Server connection
        try {
            this.mswf.releaseConnection();
        } catch (EmailAccessException e) {
            throw new SyncSourceException("Error Closing Email:" +
                                          e.getMessage(), e);
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("EmailSyncSource - end EndSync [" + sourceURI + "]");
        }

  }


  // --------------------------------------------------------- Protected Methods

  /**
   * Return the device with the given deviceId
   * @param deviceId String
   * @return Sync4jDevice
   * @throws PersistentStoreException
   */
  protected Sync4jDevice getDevice(String deviceId) throws PersistentStoreException {
      Sync4jDevice device = new Sync4jDevice(deviceId);
      PersistentStore store = Configuration.getConfiguration().getStore();
      store.read(device);
      return device;
  }


  /**
   *
   * create a single syncitem with UID and the empty content.
   * This SyncItem is used in the getDeleted operation
   *
   * @param uid String
   * @param status char
   * @return SyncItem
   */
  protected SyncItem createSyncItem(String uid, char status) {

        if (log.isLoggable(Level.FINEST)) {
          log.finest("EmailSyncSource createSyncItem start ");
        }

        SyncItem syncItem = null;
        String   stream   = "";


        syncItem = new SyncItemImpl(this, uid, status);

        if (this.encode) {
            syncItem.setContent(Base64.encode(stream.getBytes()));
        } else {
            syncItem.setContent(stream.getBytes());
        }

        if (log.isLoggable(Level.FINEST)) {
          log.finest("EmailSyncSource createSyncItem end ");
        }

        return syncItem;
  }

  /**
   * Create SyncItem from Contact
   *
   * @param mails List
   * @param status char
   * @param delete boolean
   *
   * @return SyncItem[]
   *
   * @throws EntityException
   */
  protected SyncItem[] createSyncItems(List mails,
                                         char status,
                                         boolean delete)
     throws EntityException {


        SyncItem[] syncItems  = null;
        Email      mail       = null;
        String     mailID     = null;
        int        l          = 0;

        l = mails.size();
        syncItems = new SyncItem[l];

        if (log.isLoggable(Level.FINEST)) {
          log.finest("EmailSyncSource createSyncItems items number " + l);
        }

        try {
            for (int i = 0; i < l; i++)
            {
                if (delete){
                    mailID = (String)mails.get(i);
                    syncItems[i] = createSyncItem(mailID, status);
                }else{
                    mail = (Email) mails.get(i);
                    syncItems[i] = createSyncItem(mail, status);
                }
            }
        } catch (Exception e) {
            throw new EntityException(e.getMessage());
        }

        return syncItems;
  }

  /**
   * Get Data from XML message converting the xml
   * item into a Email object
   *
   * @param content String
   * @return Email
   * @throws EntityException
   */
  protected Email getFoundationMailFromXML(String content)
            throws EntityException {

        /**
         * @todo check the replace
         */
        content = content.replaceAll("]]&gt;","]]>");

        ByteArrayInputStream  buffer = null ;
        XMLMailParser         parser = null ;
        Email                 mail   = null ;
        try {
            mail = new Email();
            buffer = new ByteArrayInputStream (content.getBytes());

            if ((content.getBytes()).length > 0) {
                parser = new XMLMailParser(buffer);
                mail = (Email) parser.parse();
            }
        } catch (Exception e){
            throw new EntityException (e.toString() );
        }
        return mail;
  }

  /**
   *
   * Get Data from sync4j.foundation.pdi.mail.Email
   * converting the Contact object into an xml item
   *
   * @param mail Email
   * @return String
   * @throws EntityException
   */
  protected String getXMLFromFoundationMail (Email mail)
       throws EntityException {

        String xml = null;
        try{
            MailToXML m2xml = new MailToXML(this.deviceTimeZone,
                                            this.deviceCharset);
            xml = m2xml.convert(mail);
        } catch (Exception e){
            throw new EntityException(e);
        }
        return xml;
  }

  /**
   *
   *
   * @param email Email
   * @param status char
   * @return SyncItem
   * @throws EntityException
   */
  protected SyncItem createSyncItem(Email email, char status)
       throws EntityException {

        SyncItem syncItem = null;
        String   stream   = null;

        stream = getXMLFromFoundationMail(email);

        /**
         * @todo check the replace
         */
        stream = "<![CDATA[" + stream + "]]>";

        String uid = (String)email.getUID().getPropertyValue();
        try {
           syncItem = new SyncItemImpl(this, uid, status);
        } catch (Exception e) {
           throw new EntityException(e);
        }

        if (this.encode) {
            syncItem.setContent(Base64.encode(stream.getBytes()));
        } else {
            syncItem.setContent(stream.getBytes());
        }

        return syncItem;
  }


  //------------------------------------------------------------ PRIVATE METHODS

  /**
   *
   * Create EntityManager istance (if not already exist)
   *
   * @return EntityManager
   * @throws EmailAccessException
   */
  private EntityManager getEntityManager() throws EmailAccessException {
      if (this.em == null) {
          this.em = new EntityManager(this.mswf);
      }
      return this.em;
  }


  /**
   * set the filter
   *
   *  @todo should be a recursive algorithm.
   *  At the moment I do a fix parser (only 2 level)
   *
   * @param logicalClause LogicalClause
   */
  private void setFilter(LogicalClause logicalClause)  {

     if (log.isLoggable(Level.FINEST)) {
        log.finest("\nFilter clause " + logicalClause );
     }
     Clause[] operands = logicalClause.getOperands();

     for (int i=0; i<operands.length; i++){

         if        (operands[i] instanceof IDClause) {
             IDClause tmp = (IDClause)operands[i];
             setIdFilter(tmp);
         } else if (operands[i] instanceof WhereClause) {
             WhereClause tmp = (WhereClause)operands[i];
             setTimeFilter(tmp);
         } else if (operands[i] instanceof FieldClause) {
             FieldClause tmp = (FieldClause)operands[i];
             setSizeFilter(tmp);
         } else if (operands[i] instanceof LogicalClause) {

             LogicalClause tmp = (LogicalClause)operands[i];
             Clause[] ops = tmp.getOperands();

             for (int y=0; y<ops.length; y++){
                if         (ops[y] instanceof IDClause){
                    IDClause tmp2 = (IDClause)ops[y];
                    setIdFilter(tmp2);
                } else  if (ops[y] instanceof WhereClause) {
                    WhereClause tmp2 = (WhereClause)ops[y];
                    setTimeFilter(tmp2);
                } else if (ops[y] instanceof FieldClause){
                    FieldClause tmp2 = (FieldClause)ops[y];
                    setSizeFilter(tmp2);
                }
             }

         }
     }

  }

  /**
   *
   * set the time filter
   *
   * @param clause WhereClause
   */
  private void setTimeFilter(WhereClause clause)  {

     if (log.isLoggable(Level.FINEST)) {
        log.finest("setting setTimeFilter ");
     }

     String property = clause.getProperty();
     String operator = clause.getOperator();

     if (property.equalsIgnoreCase(Def.FILTER_TIME_FIELD)){

        String[] values = clause.getValues();
        String value = values[0];

        if (log.isLoggable(Level.FINEST)) {
           log.finest("operator " + operator );
           log.finest("property " + property);
           log.finest("value    " + value);
        }

        if (value != null){
           if (value.equalsIgnoreCase("ALL")){
             this.timeFilter = null;
           } else {
             this.timeFilter = Utility.UtcToDate(value);
           }
        }
     }
  }



  /**
   *
   * set the time filter
   *
   *
   * @param icIdFilter IDClause
   */
  private void setIdFilter(IDClause icIdFilter)  {

     if (log.isLoggable(Level.FINEST)) {
        log.finest("setting setIdFilter ");
     }

     String property = icIdFilter.getProperty();
     String operator = icIdFilter.getOperator();

     String[] values = icIdFilter.getValues();
     String value = values[0];

     if (log.isLoggable(Level.FINEST)) {
         log.finest("operator " + operator );
         log.finest("property " + property);
         log.finest("value    " + value);
     }

     this.idFilter = value;


  }



  /**
   *
   * set the time filter
   *
   * @param fcSizeFilter FieldClause
   */
  private void setSizeFilter(FieldClause fcSizeFilter)  {

     if (log.isLoggable(Level.FINEST)) {
        log.finest("setting setSizeFilter ");
     }

     Property property = fcSizeFilter.getProperty();

     String moduleName = null;
     int    maxSize    = 0;

     moduleName = property.getPropName();
     maxSize    = property.getMaxSize();

     if (log.isLoggable(Level.FINEST)) {
        log.finest(" FieldClause module: " + moduleName + " size: " + maxSize);
     }

     if (moduleName.equalsIgnoreCase(Def.FILTER_SIZE_LABEL_HEADER)){

        if        ((maxSize == 0) && (!isBodyFieldSet) && (!isAttachFieldSet)) {
            this.sizeFilter = 1;
        } else if ((maxSize > 0)  && (!isBodyFieldSet) && (!isAttachFieldSet)) {
            this.sizeFilter = 3;
            this.numBytes   = maxSize;
        }
        this.isHeaderFieldSet = true;

     } else if (moduleName.equalsIgnoreCase(Def.FILTER_SIZE_LABEL_BODY)){

        if        ((maxSize == 0) && (!isAttachFieldSet)){
            this.sizeFilter = 5;
            this.numBytes   = 0;
        } else if ((maxSize > 0)  && (!isAttachFieldSet)){
            this.sizeFilter = 15;
            this.numBytes   = maxSize;
        }
        this.isBodyFieldSet = true;

     } else if (moduleName.equalsIgnoreCase(Def.FILTER_SIZE_LABEL_ATTACH)){

        if        (maxSize == 0){
            this.sizeFilter = 21;
            this.numBytes   = 0;
        } else if (maxSize > 0) {
            this.sizeFilter = 63;
            this.numBytes   = maxSize;
        }
        this.isAttachFieldSet = true;

     }


  }

}
