/**
 * Copyright (C) 2005-2006 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.email.items.manager ;


import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.Locale;
import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;

import javax.mail.Message;
import javax.mail.Session;
import javax.mail.Folder;
import javax.mail.Flags;
import javax.mail.Transport;
import com.sun.mail.imap.IMAPMessage;
import com.sun.mail.imap.IMAPFolder;
import javax.mail.internet.MimeMessage;
import javax.mail.Part;
import javax.mail.Multipart;
import javax.mail.BodyPart;
import com.sun.mail.imap.IMAPBodyPart;
import javax.mail.MessagingException;
import javax.mail.internet.ContentType;

import com.funambol.email.pdi.mail.Email;
import com.funambol.email.pdi.mail.EmailItem;
import com.funambol.email.items.dao.EntityDAO;
import com.funambol.email.items.model.ClientMessage;
import com.funambol.email.transport.IMailServerWrapper;
import com.funambol.email.transport.CommonMailServerWrapper;
import com.funambol.email.transport.ImapMailServerWrapper;
import com.funambol.email.transport.PopMailServerWrapper;
import com.funambol.email.util.Def;
import com.funambol.email.util.Utility;
import com.funambol.email.exception.*;



/**
 *
 * @version $Id: EntityManager.java,v 1.3 2006/02/06 09:02:15 gbmiglia Exp $
 */
public class EntityManager {


    // ------------------------------------------------------------ Private data

    private boolean stop = false;



    // -------------------------------------------------------------- Properties

    /**     */
    CommonMailServerWrapper cmsw = null;

    /**     */
    protected Logger  log = Logger.getLogger(Def.LOGGER_NAME);

    /**     */
    public EntityDAO ed       = null;

    /**     */
    long timeStart = 0;
    long timeStop  = 0;

    // ------------------------------------------------------------ Costructors


    /**
     *
     */
    public EntityManager(){
    }

    /**
     *
     *
     *
     * @param msw_e MailServerWrapper
     * @throws EmailAccessException
     */

    public EntityManager(IMailServerWrapper msw_e)
     throws EmailAccessException {
       cmsw = (CommonMailServerWrapper)msw_e;
       ed   = new EntityDAO();
    }

   // ----------------------------------------------------------- Public Methods



   /**
    * deletes an item
    *
    *
    * @param FMID String
    * @param msw IMailServerWrapper
    * @throws EntityException
    */
   public void moveSentItem(String FMID,
                            IMailServerWrapper msw)
      throws EntityException {

        if (msw instanceof ImapMailServerWrapper){

           // imap protocol

           ImapMailServerWrapper imsw = (ImapMailServerWrapper)msw;

           IMAPFolder sourceFolder = imsw.getOutboxFolder();
           IMAPFolder targetFolder = imsw.getSentFolder();

           try {

             timeStart = System.currentTimeMillis();
             sourceFolder.open(Folder.READ_WRITE);
             ed.moveImapItem(sourceFolder,
                               targetFolder,
                               Long.parseLong(FMID));

           }catch (MessagingException me){
               throw new EntityException(me);
           }catch (EntityException e){
              throw new EntityException(e);
           } finally {
              if (sourceFolder != null){
                try {
                  if (sourceFolder.isOpen()){
                     sourceFolder.close(true);
                     timeStop = System.currentTimeMillis();
                     if (log.isLoggable(Level.FINEST)) {
                        log.finest("moveSentItem Execution Time: " +
                                   (timeStop - timeStart) + " ms");
                     }
                  }
                } catch (MessagingException me){
                  throw new EntityException(me);
                }
              }
           }

        } else if (msw instanceof PopMailServerWrapper){

           // pop protocol

           PopMailServerWrapper pmsw = (PopMailServerWrapper)msw;

           /**
            * pop protocol can't access to OUTBOX
            */

        }


   }


    /**
     *
     *
     * @param FID String
     * @param emailNew Email
     * @param msw IMailServerWrapper
     * @throws EntityException
     */
    public void sendItem(String FID,
                         Email emailNew,
                         IMailServerWrapper msw)
      throws EntityException {

        if (log.isLoggable(Level.FINEST)) {
          log.finest("EntityManager sendItem start");
        }

        Message msg = null;

        if (msw instanceof ImapMailServerWrapper){
           ImapMailServerWrapper imsw = (ImapMailServerWrapper)msw;
           msg = createMessage(imsw.getSession(), emailNew, FID);
        } else if (msw instanceof PopMailServerWrapper){
           PopMailServerWrapper pmsw = (PopMailServerWrapper)msw;
           msg = createMessage(pmsw.getSession(), emailNew, FID);
        }


        // send message
        try {
             timeStart = System.currentTimeMillis();
             Transport.send(msg);
             timeStop = System.currentTimeMillis();
             if (log.isLoggable(Level.FINEST)) {
                 log.finest("insertItem (send) Execution Time: " +
                            (timeStop - timeStart) + " ms");
             }
        }catch (MessagingException me){
           throw new EntityException(me);
        }

        if (log.isLoggable(Level.FINEST)) {
          log.finest("EntityManager sendItem end");
        }


    }


   // -------------------------------------------------------- Protected Methods


    /**
     * create a javaMail Message from a funambol foundation Email
     *
     *
     * @param s Session
     * @param ctmp Email
     * @param FID String
     * @return Message
     * @throws EntityException
     */
    public Message createMessage(Session s, Email ctmp, String FID)
      throws EntityException {

        Message doc = null;
        try {

            EmailItem emailItem = ctmp.getEmailItem();

            String rfc2822 = (String)emailItem.getStringRFC2822().getPropertyValue();
            byte[] b = rfc2822.getBytes();
            InputStream is = new ByteArrayInputStream(b);

            doc = new MimeMessage(s, is);

            // if draft set the FLAG
            if (FID.equalsIgnoreCase(Def.FOLDER_DRAFTS_ID)){
               doc.setFlag(Flags.Flag.DRAFT, true);
            }

            // we must set the message as "sync message"
            doc.setHeader(Def.HEADER_LABEL_SYNCMESSAGE,
                          Def.HEADER_VALUE_SYNCMESSAGE);

            /**
             * @todo check if correct
             */
            // get the date (convert bb_7290 if needed)
            //java.util.Date originationDate = Utility.getOriginationDate(doc, this.cmsw.getLocale());
            //doc.setHeader(Def.HEADER_LABEL_DATE, ""+originationDate);



        } catch (Exception e){
            throw new EntityException(e);
        }

        if (log.isLoggable(Level.FINEST)) {
          log.finest("end creation Message");
        }

        return doc;
    }

    /**
     *
     *
     * @param smsg ServerMessage
     * @param loc Locale
     * @return Email
     * @throws EntityException
     */
    public Email createFoundationMail(ClientMessage smsg, Locale loc)
     throws EntityException {

        Email mtmp       = new Email();
        String uid       = null;
        String created   = null; // Date created
        String received  = null; // Date received
        String modified  = null; // Date modified
        String read      = null; // boolean read
        String forwarded = null; // boolean forwarded
        String replied   = null; // boolean replied
        String deleted   = null; // boolean deleted
        String flagged   = null; // boolean flagged
        String emailItem = null;

        try {

           uid       = smsg.getUID();
           created   = smsg.getCreated();
           received  = smsg.getReceived();
           modified  = smsg.getModified();
           read      = smsg.getRead();
           forwarded = smsg.getForwarded();
           replied   = smsg.getReplied();
           deleted   = smsg.getDeleted();
           flagged   = smsg.getFlagged();
           emailItem = smsg.getStreamMessage();

        } catch (Exception e) {
           e.printStackTrace();
        }

        mtmp.getUID().setPropertyValue(uid);

        if (created != null){
          mtmp.getCreated().setPropertyValue(created);
        }
        if (received != null){
          mtmp.getReceived().setPropertyValue(received);
        }
        if (modified != null){
          mtmp.getModified().setPropertyValue(modified);
        }

        if (read != null){
          mtmp.getRead().setPropertyValue(read);
        }
        if (forwarded != null){
          mtmp.getForwarded().setPropertyValue(forwarded);
        }
        if (replied != null){
          mtmp.getReplied().setPropertyValue(replied);
        }
        if (deleted != null){
          mtmp.getDeleted().setPropertyValue(deleted);
        }
        if (flagged != null){
          mtmp.getFlagged().setPropertyValue(flagged);
        }

        mtmp.getEmailItem().getStringRFC2822().setPropertyValue(emailItem);

        return mtmp;
    }


    /**
     *
     *
     * @param multipart Multipart
     * @param numBytes int
     * @return Object[]
     * @throws EntityException
     */
    protected Object[] handleMultipart(Multipart multipart, int numBytes)
      throws EntityException {

      Object[] out = new Object[2];

      stop = false;

      try {

         int n = multipart.getCount();
         if (log.isLoggable(Level.FINEST)) {
            log.finest("handleMultipart " + n + " part! ");
         }

         for (int i=0; i<n; i++) {
            if (log.isLoggable(Level.FINEST)) {
               log.finest("start analize the " + i + " part ");
            }
            out = handlePart(multipart.getBodyPart(i), numBytes);
            if (log.isLoggable(Level.FINEST)) {
               log.finest("stop " + stop);
            }
            if (stop){
               break;
            }
         }
      } catch (MessagingException me){
         throw new EntityException(me);
      }
      return out;
    }


    /**
     * retruns the body of the mail without attachment.
     *
     * output
     *
     * [0] ()
     *  - string if we have a text/plain
     *  - html   if we have a text/html
     *  - empty string otherwise
     *
     * [1] (String)
     *  - related content type
     *
     * if the size part is greater than "numbytes" the methods
     * returns the default content
     *
     * @todo
     *  - manage UTF7 :"UTF7"
     *                 "UNICODE-1-1-UTF-7",
     *                 "csUnicode11UTF7",
     *                 "UNICODE-2-0-UTF-7"
     *
     * @param part Part
     * @param numBytes int
     * @return Object[]
     * @throws EntityException
     */
    protected Object[] handlePart(Part part, int numBytes)
      throws EntityException {

       Object[] out = new Object[2];
       // default
       out[0] = Def.CONTENT_VALUE_BODY;
       out[1] = Def.CONTENT_VALUE_CONTENTTYPE;

       try {
          String contentType     = null;

          String disposition     = part.getDisposition();
          String contentTypeOrig = part.getContentType().toLowerCase();
          int    size            = part.getSize();

          if (log.isLoggable(Level.FINEST)) {
            log.finest("\noriginal content type: " + contentTypeOrig +
                       "\ndisposition: " + disposition +
                       "\nsize:        " + size +
                       "\nnumBytes:    " + numBytes);
          }

          if ((numBytes == -1) || (size < numBytes)) {
             // get content
          } else {
             // if numBytes<size get nothing
             return out;
          }

          /**
           * @todo
           */
          if (contentTypeOrig.equalsIgnoreCase("message/delivery-status") ||
              contentTypeOrig.equalsIgnoreCase("message/rfc822") ){
              return out;
          }

          if ( contentTypeOrig.length() >= 10 ){
             int index   = contentTypeOrig.indexOf(";");
             contentType = contentTypeOrig.toLowerCase().substring(0, index);
          }

          ContentType ct = new ContentType(part.getContentType());
          String specifiedCharset = ct.getParameter("charset");

          if (log.isLoggable(Level.FINEST)) {
            log.finest("\nContent Type Orig : " + contentTypeOrig +
                       "\nContent Type New  : " + contentType +
                       "\nDisposition       : " + disposition +
                       "\nAttachment        : " + part.ATTACHMENT +
                       "\nspecifiedCharset  : " + specifiedCharset +
                       "\nContent Class     : " + part.getClass());
          }

          //
          if (part instanceof IMAPBodyPart ||
              part instanceof IMAPMessage ||
              part instanceof Message) {

             if (disposition == null){

                if (contentType.equals("text/plain")) {

                   if ("utf-7".equalsIgnoreCase(specifiedCharset) ||
                       "unicode-1-1-utf-7".equalsIgnoreCase(specifiedCharset) ){
                       out[0] = Def.CONTENT_VALUE_BODY;
                   } else {
                       out[0] = (String)part.getContent();
                   }
                   out[1] = contentTypeOrig;
                   stop = true;

                } else if (contentType.equals("text/html")) {

                   if ("utf-7".equalsIgnoreCase(specifiedCharset) ||
                       "unicode-1-1-utf-7".equalsIgnoreCase(specifiedCharset) ){
                       out[0] = Def.CONTENT_VALUE_BODY;
                   } else {
                       out[0] = (String)part.getContent();
                   }
                   out[1] = contentTypeOrig;
                   stop = true;

                } else {
                  // do nothing
                  stop = false;
                }

             } else if (disposition.equalsIgnoreCase(Part.ATTACHMENT)) {
                // do nothing
                stop = false;
             } else if (disposition.equalsIgnoreCase(Part.INLINE)) {
                // do nothing
                stop = false;
             } else {
                // do nothing
                stop = false;
             }

          } else {
             // do nothing
             stop = false;
          }

       } catch (MessagingException me){
          throw new EntityException(me);
       } catch (IOException ioe){
          throw new EntityException(ioe);
       }


       return out;
    }

}
