/**
 * Copyright (C) 2005-2006 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.email.items.manager ;


import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.Date;
import java.sql.Timestamp;
import java.util.Enumeration;
import java.io.IOException;

import javax.mail.Message;
import javax.mail.internet.MimeMessage;
import javax.mail.Header;
import javax.mail.Multipart;
import javax.mail.Flags;
import javax.mail.MessagingException;
import javax.mail.Folder;
import com.sun.mail.pop3.POP3Folder;

import com.funambol.email.pdi.mail.Email;
import com.funambol.email.items.dao.PopEntityDAO;
import com.funambol.email.items.model.ClientMessage;
import com.funambol.email.items.model.EISMessage;
import com.funambol.email.transport.IMailServerWrapper;
import com.funambol.email.transport.PopMailServerWrapper;
import com.funambol.email.util.Def;
import com.funambol.email.util.Utility;
import com.funambol.email.cache.CrcSyncItemInfo;
import com.funambol.email.exception.*;


/**
 *
 * @version $Id: PopEntityManager.java,v 1.2 2006/02/06 09:02:15 gbmiglia Exp $
 */
public class PopEntityManager extends EntityManager {


    // -------------------------------------------------------------- Properties

    /**     */
    public PopMailServerWrapper pmsw = null;

    /**     */
    public PopEntityDAO ped       = null;


    // ------------------------------------------------------------ Costructors


    /**
     *
     */
    public PopEntityManager(){
    }

   /**
    *
    *
    *
    * @param msw_e IMailServerWrapper
    * @throws EmailAccessException
    */
   public PopEntityManager(IMailServerWrapper msw_e)
     throws EmailAccessException {
        pmsw = (PopMailServerWrapper)msw_e;
        ped  = new PopEntityDAO();
    }


    // ---------------------------------------------------------- Public Methods

   /**
    * deletes an item
    *
    *
    * @param FID String
    * @param FMID String
    * @throws EntityException
    */
   public void removeItem(String FID, String FMID)
     throws EntityException {

        POP3Folder f = this.pmsw.getInboxFolder();

        try {

             timeStart = System.currentTimeMillis();
             f.open(Folder.READ_WRITE);
             ped.removeItem(f, FMID);

        }catch (MessagingException me){
           throw new EntityException(me);
        }catch (EntityException e){
           throw new EntityException(e);
        } finally {
           if (f != null){
             try {
                if (f.isOpen()){
                   f.close(true);
                   timeStop = System.currentTimeMillis();
                   if (log.isLoggable(Level.FINEST)) {
                      log.finest("removeItem Execution Time: " +
                                (timeStop - timeStart) + " ms");
                   }
                }
             } catch (MessagingException me){
                throw new EntityException(me);
             }
           }
        }
   }



    /**
     *
     *
     * @param folderFilter int
     * @param idFilter String
     * @param timeFilter Date
     * @return String[]
     * @throws EntityException
     */
    public String[] getAllItems(int folderFilter,
                                String idFilter,
                                Date timeFilter)
      throws EntityException {

        if (log.isLoggable(Level.FINEST)) {
          log.finest("Start - getAllItems() ");
        }

        // output
        String[] allitems =  null;

        // always get inbox
        allitems  = this.getItems(this.pmsw.getInboxFolder(),
                                  idFilter,
                                  timeFilter);

        if (log.isLoggable(Level.FINEST)) {
          log.finest("End - getAllItems() ");
        }

        //return allMails;
        return allitems;
    }

    /**
     *
     *
     *
     * @param folderFilter int
     * @param idFilter String
     * @param timeFilter Date
     * @return CrcSyncItemInfo[]
     * @throws EntityException
     */
    public CrcSyncItemInfo[] getAllSyncItemInfo(int folderFilter,
                                                String idFilter,
                                                Date timeFilter)
      throws EntityException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("getAllSyncItemInfo Start");
        }

        // all
        CrcSyncItemInfo[] syncItemInfos = null;

        // always get inbox
        syncItemInfos  = this.getItemsInfo(this.pmsw.getInboxFolder(),
                                           idFilter,
                                           timeFilter);


        if (log.isLoggable(Level.FINEST)) {
           log.finest("getAllSyncItemInfo List: "+syncItemInfos.length+" Document");
        }

        return  syncItemInfos;

    }

    /**
     *
     *
     * @param folderFilter int
     * @param timeFilter Date
     * @param sizeFilter int
     * @param numBytes int
     * @param UID String
     * @return Email
     * @throws EntityException
     */
    public Email getItemByUID(int folderFilter,
                              Date timeFilter,
                              int sizeFilter,
                              int numBytes,
                              String UID)
    throws EntityException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("getItemByUID start " + UID);
        }

        // create the Mail Server UID and Mail server Folder ID
        String FID  = Utility.getFID(UID);
        String FMID = Utility.getFMID(UID);
        if (log.isLoggable(Level.FINEST)) {
            log.finest("getItemByUID folderId: " + FID + " uid " + FMID );
        }

        // get the right folder according to Folder ID
        POP3Folder f = Utility.getFolderPOP3(FID, this.pmsw);

        // get Message
        ClientMessage clientMsg = this.getItem(f,
                                               FMID,
                                               sizeFilter,
                                               numBytes);

        // convert ServerMessage in Email
        Email emailOut = null;
        if (clientMsg != null){
            try {
                emailOut = createFoundationMail(clientMsg,
                                                this.pmsw.getLocale());
            } catch (Exception e) {
                throw new EntityException(e);
            }
        }

        return emailOut;

    }




    /**
     * checks if a contact is already inserted in the db
     * and returns the Document in the DB
     * we use firstname, lastname as search criteria
     *
     *
     * @param FID String
     * @param FMID String
     * @param ctmp Email
     * @return Email
     * @throws EntityException
     */
    public Email getItemByWhereClause(String FID, String FMID, Email ctmp)
      throws EntityException {

       if (log.isLoggable(Level.FINEST)) {
            log.finest("getItemByWhereClause start ");
       }


       POP3Folder f        = Utility.getFolderPOP3(FID, this.pmsw);

       String messageIdExt = null;
       String messageIdInt = null;
       ClientMessage found = null;
       Email ctmpout       = null;

       try {

          Message msgExt  = createMessage(this.pmsw.getSession(), ctmp, FID);
          messageIdExt    = Utility.getMessageID(msgExt);

          timeStart = System.currentTimeMillis();
          f.open(Folder.READ_WRITE);

          if (messageIdExt != null){

             //
             EISMessage msgInt = ped.getItemByMessageID(f,
                                                        FID,
                                                        messageIdExt);
             // check
             if (msgInt != null){
                if (log.isLoggable(Level.FINEST)) {
                  log.finest("Email already Exist");
                }
                found = setClientMessage(f,
                                         msgInt.getGUID(),
                                         msgInt.getMessage());
             }
          }

       } catch (Exception e){
          throw new EntityException(e);
       } finally {
          if (f != null){
             try {
                if (f.isOpen()){
                   f.close(true);
                   timeStop = System.currentTimeMillis();
                   if (log.isLoggable(Level.FINEST)) {
                      log.finest("getItemByWhereClause Execution Time: " +
                                (timeStop - timeStart) + " ms");
                   }
                }
             } catch (MessagingException me){
               throw new EntityException(me);
             }
          }
       }


       // convert Message in a Email
       if (found != null){
           try {
               ctmpout = createFoundationMail(found, this.pmsw.getLocale());
           } catch (Exception e) {
               throw new EntityException(e);
           }
       }

        return ctmpout;
    }

    /**
     * returns Header
     *
     * @param f POP3Folder
     * @param msg Message
     * @return ClientMessage
     * @throws EntityException
     */
    private ClientMessage getPart_H(POP3Folder f,
                                    Message msg)
      throws EntityException {

        ClientMessage clientMsg = null;

        try {

           Message newMsg = new MimeMessage(this.pmsw.getSession());

           // header
           String mid = Utility.setNewHeader(newMsg, msg);

           // default content
           Object[] newContent = new Object[2];
           newContent[0] = Def.CONTENT_VALUE_BODY;
           newContent[1] = Def.CONTENT_VALUE_CONTENTTYPE;

           newMsg.setContent(newContent[0], String.valueOf(newContent[1]));

           newMsg.saveChanges();

           // set the new message wtih the old Message-ID
           newMsg.setHeader(Def.HEADER_LABEL_MESSAGE_ID, mid);

           // create UID i.e. I/45
           String FID  = Utility.getFID(f, this.pmsw);
           String FMID = f.getUID(msg);
           String UID  = FID + Def.SEPARATOR + FMID ;

           clientMsg = this.setClientMessage(f , UID, newMsg);

        } catch (MessagingException me){
           throw new EntityException(me);
        }

        return clientMsg;

    }

    /**
     *
     * @todo:
     * returns only N bytes of the header
     * we must understand ths filter
     *
     * @param f POP3Folder
     * @param msg Message
     * @param numBytes int
     * @return ClientMessage
     * @throws EntityException
     */
    private ClientMessage getPart_H(POP3Folder f,
                                    Message msg,
                                    int numBytes)
      throws EntityException {

        ClientMessage clientMsg = null;

        try {

           Message newMsg = new MimeMessage(this.pmsw.getSession());

           // header
           String mid = Utility.setNewHeader(newMsg, msg);

           // default content
           Object[] newContent = new Object[2];
           newContent[0] = Def.CONTENT_VALUE_BODY;
           newContent[1] = Def.CONTENT_VALUE_CONTENTTYPE;

           newMsg.setContent(newContent[0], String.valueOf(newContent[1]));

           newMsg.saveChanges();

           // set the new message wtih the old Message-ID
           newMsg.setHeader(Def.HEADER_LABEL_MESSAGE_ID,mid);

           // create UID i.e. I/45
           String FID  = Utility.getFID(f, this.pmsw);
           String FMID = f.getUID(msg);
           String UID  = FID + Def.SEPARATOR + FMID ;

           clientMsg = this.setClientMessage(f , UID, newMsg);

        } catch (MessagingException me){
           throw new EntityException(me);
        }

        return clientMsg;

    }

    /**
     *
     * returns Header and Body
     *
     * @param f POP3Folder
     * @param msg Message
     * @return ClientMessage
     * @throws EntityException
     */
    private ClientMessage getPart_H_B(POP3Folder f,
                                      Message msg)
      throws EntityException {

        ClientMessage clientMsg = null;

        try {

           Message newMsg = new MimeMessage(this.pmsw.getSession());

           // header
           String mid = Utility.setNewHeader(newMsg, msg);

           // content
           Object content = msg.getDataHandler().getContent();
           Object[] newContent = new Object[2];
           if (content instanceof Multipart) {
              newContent = handleMultipart((Multipart)content, -1);
           } else {
              newContent = handlePart(msg, -1);
           }

           newMsg.setContent(newContent[0], String.valueOf(newContent[1]));

           newMsg.saveChanges();

           // set the new message wtih the old Message-ID
           newMsg.setHeader(Def.HEADER_LABEL_MESSAGE_ID,mid);

           // create UID i.e. I/45
           String FID  = Utility.getFID(f, this.pmsw);
           String FMID = f.getUID(msg);
           String UID  = FID + Def.SEPARATOR + FMID ;

           clientMsg = this.setClientMessage(f , UID, newMsg);

        } catch (MessagingException me){
           throw new EntityException(me);
        } catch (IOException ioe){
           throw new EntityException(ioe);
        }

        return clientMsg;

    }

    /**
     * returns Header and N bytes of the Body
     *
     * @param f POP3Folder
     * @param msg Message
     * @param numBytes int
     * @return ClientMessage
     * @throws EntityException
     */
    private ClientMessage getPart_H_B(POP3Folder f,
                                      Message msg,
                                      int numBytes)
      throws EntityException {

        ClientMessage clientMsg = null;

        try {

           Message newMsg = new MimeMessage(this.pmsw.getSession());

           // header
           String mid = Utility.setNewHeader(newMsg, msg);

           // content
           Object content = msg.getDataHandler().getContent();
           Object[] newContent = new Object[2];
           if (content instanceof Multipart) {
              newContent = handleMultipart((Multipart)content, numBytes);
           } else {
              newContent = handlePart(msg, numBytes);
           }

           newMsg.setContent(newContent[0], String.valueOf(newContent[1]));

           newMsg.saveChanges();

           // set the new message wtih the old Message-ID
           newMsg.setHeader(Def.HEADER_LABEL_MESSAGE_ID,mid);

           // create UID i.e. I/45
           String FID  = Utility.getFID(f, this.pmsw);
           String FMID = f.getUID(msg);
           String UID  = FID + Def.SEPARATOR + FMID ;

           clientMsg = this.setClientMessage(f , UID, newMsg);

        } catch (MessagingException me){
           throw new EntityException(me);
        } catch (IOException ioe){
           throw new EntityException(ioe);
        }

        return clientMsg;

    }


    /**
     * returns all parts of the message
     *
     * @param f POP3Folder
     * @param msg Message
     * @return ClientMessage
     * @throws EntityException
     */
    private ClientMessage getPart_H_B_A(POP3Folder f,
                                       Message msg)
      throws EntityException {
        ClientMessage clientMsg;
        try {

           // create UID i.e. I/45
           String FID  = Utility.getFID(f, this.pmsw);
           String FMID = f.getUID(msg);
           String UID  = FID + Def.SEPARATOR + FMID ;

           clientMsg = this.setClientMessage(f, UID, msg);

        } catch (MessagingException me){
           throw new EntityException(me);
        }
        return clientMsg;
    }


    /**
     * returns Header, Body and N bytes of the Attach
     *
     * @param f POP3Folder
     * @param msg Message
     * @param numBytes int
     * @return ClientMessage
     * @throws EntityException
     */
    private ClientMessage getPart_H_B_A(POP3Folder f,
                                        Message msg,
                                        int numBytes)
      throws EntityException {
        ClientMessage clientMsg;
        try {

           // create UID i.e. I/45
           String FID  = Utility.getFID(f, this.pmsw);
           String FMID = f.getUID(msg);
           String UID  = FID + Def.SEPARATOR + FMID ;

           clientMsg = this.setClientMessage(f, UID, msg);

        } catch (MessagingException me){
           throw new EntityException(me);
        }
        return clientMsg;
    }

    /**
     * convert Message into ServerMessages
     *
     *
     *
     * @param f POP3Folder
     * @param UID String
     * @param msg Message
     * @return ClientMessage
     * @throws EntityException
     */
    private ClientMessage setClientMessage(POP3Folder f,
                                           String UID,
                                           Message msg)
      throws EntityException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("convertMessageToClientMessage() start " + f.getName());
        }

        ClientMessage clientMsg  = null;

        String        message_id = null;
        String        stream     = null;

        String        created    = null;
        String        received   = null;
        String        modified   = null;

        String        read       = null;
        String        forwarded  = null;
        String        replied    = null;
        String        deleted    = null;
        String        flagged    = null;

        try {

            // header message-id
            message_id = Utility.getMessageID(msg);

            // message
            stream = Utility.messageToString(msg);

            //origination date
            created = Utility.DateToUTC(
                        Utility.getOriginationDate(msg, this.pmsw.getLocale()),
                        this.pmsw.getLocale());
            //received date
            received = Utility.getReceivedDate(msg, this.pmsw.getLocale());
            //modified date
            modified = null;

            // read
            read = String.valueOf(msg.isSet(Flags.Flag.SEEN));
            // forwarded
            forwarded = null;
            // replied
            replied = null;
            // deleted
            deleted = String.valueOf(msg.isSet(Flags.Flag.DELETED));
            // flagged
            flagged = String.valueOf(msg.isSet(Flags.Flag.FLAGGED));

            clientMsg = new ClientMessage(UID,
                                          message_id,
                                          stream,
                                          created,
                                          received,
                                          modified,
                                          read,
                                          forwarded,
                                          replied,
                                          deleted,
                                          flagged);

        } catch (MessagingException e){
            throw new EntityException(e);
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("convertMessageToServerMessage() - ServerMessage created ");
        }

        return clientMsg;

    }

    /**
     *
     *
     * @param f POP3Folder
     * @param uid String
     * @param sizeFilter int
     * @param numbytes int
     *
     * @return ClientMessage
     *
     * @throws EntityException
     */
    private ClientMessage getItem(POP3Folder f,
                                  String uid,
                                  int sizeFilter,
                                  int numbytes)
       throws EntityException {

        ClientMessage clientMsg = null;

        try {
             timeStart = System.currentTimeMillis();
             f.open(Folder.READ_WRITE);
             // get message
             Message msg = ped.getItemByUID(f, uid);
             // filter the original message
             clientMsg = this.getPart(f, sizeFilter, numbytes, msg);
        } catch (Exception e){
            throw new EntityException(e);
        } finally {
            if (f != null){
              try {
                if (f.isOpen()){
                   f.close(true);
                   timeStop = System.currentTimeMillis();
                   if (log.isLoggable(Level.FINEST)) {
                      log.finest("getItemByUID Execution Time: " +
                                 (timeStop - timeStart) +  " ms");
                   }
                }
              } catch (MessagingException me){
                throw new EntityException(me);
              }
            }
        }

        return clientMsg;
    }


    /**
     *  get messages from folders
     *
     *
     *
     * @param f POP3Folder
     * @param idFilter String
     * @param timeFilter Date
     * @return String[]
     * @throws EntityException
     */
    private String[] getItems(POP3Folder f,
                              String idFilter,
                              Date timeFilter)
      throws EntityException {

       String[] msgIds = null;

       try {

           timeStart = System.currentTimeMillis();
           f.open(Folder.READ_WRITE);
           Message[] messages = ped.getAllItems(f, idFilter, timeFilter);

           String FID = Utility.getFID(f, this.pmsw);
           if (messages != null){
              int numMsgs = messages.length;
              msgIds      = new String[numMsgs];
              String FMID = null;

              for (int i=0; i<numMsgs; i++){
                 FMID = String.valueOf(f.getUID(messages[i]));
                 msgIds[i] = FID + Def.SEPARATOR + FMID;
              }
           }

       } catch (MessagingException e){
           throw new EntityException(e);
       } finally {
           if (f != null){
             try {
                if (f.isOpen()){
                   f.close(true);
                   timeStop = System.currentTimeMillis();
                   if (log.isLoggable(Level.FINEST)) {
                      log.finest("getItems Execution Time: " +
                                (timeStop - timeStart) + " ms");
                   }
                }
             } catch (MessagingException me){
                throw new EntityException(me);
             }
           }
       }

       return msgIds;

    }

    /**
     *  get messages from folders
     *
     *
     * @param f POP3Folder
     * @param idFilter String
     * @param timeFilter Date
     * @return CrcSyncItemInfo[]
     * @throws EntityException
     */
    private CrcSyncItemInfo[] getItemsInfo(POP3Folder f,
                                           String idFilter,
                                           Date timeFilter)
      throws EntityException {

       List            allSyncItemInfo = new ArrayList();
       CrcSyncItemInfo syncItemInfo    = null;

       try {

           timeStart = System.currentTimeMillis();
           f.open(Folder.READ_WRITE);

           // get message
           Message[] items = ped.getAllItems(f, idFilter, timeFilter);

           Message msg     = null;
           String key      = null;
           long   lastCrc  = 0;
           int    itemsNum = 0;

           if (items != null){
              itemsNum = items.length;
           }

           for (int i = 0; i < itemsNum; i++){

               msg = items[i];

               // last modification
               lastCrc  = Utility.setLastCrc(msg);

               // make id
               String FID  = Def.FOLDER_INBOX_ID;
               String FMID = f.getUID(msg);
               key         = FID + Def.SEPARATOR + FMID ;

               // create info
               syncItemInfo = new CrcSyncItemInfo();

               if (log.isLoggable(Level.FINEST)) {
                  log.finest("Pop3Email getItemsInfo info: " +
                             " uid " + key + " crc " + lastCrc);
               }

               syncItemInfo.setLastCrc(lastCrc);
               syncItemInfo.setKey(key);

               allSyncItemInfo.add(syncItemInfo);

           }

       } catch (MessagingException e){
           throw new EntityException(e);
       } finally {
           if (f != null){
             try {
                if (f.isOpen()){
                   f.close(true);
                   timeStop = System.currentTimeMillis();
                   if (log.isLoggable(Level.FINEST)) {
                      log.finest("getItemsInfo Execution Time: " +
                                 (timeStop - timeStart) + " ms");
                   }
                }
             } catch (MessagingException me){
                throw new EntityException(me);
             }
           }
       }

       return (CrcSyncItemInfo[])allSyncItemInfo.toArray(new CrcSyncItemInfo[0]);

    }

    /**
     *
     *
     * @param f POP3Folder
     * @param sizeFilter int
     * @param numBytes int
     * @param msg Message
     * @return ClientMessage
     * @throws EntityException
     */
    private ClientMessage getPart(POP3Folder f,
                                  int sizeFilter,
                                  int numBytes,
                                  Message msg)
      throws EntityException {

       ClientMessage clientMsg = null;
       if (msg != null){
          if        (sizeFilter == Def.FILTER_SIZE_H){
             clientMsg = getPart_H(f, msg);
          } else if (sizeFilter == Def.FILTER_SIZE_HPERC){
             clientMsg = getPart_H(f, msg, numBytes);
          } else if (sizeFilter == Def.FILTER_SIZE_H_B){
             clientMsg = getPart_H_B(f, msg);
          } else if (sizeFilter == Def.FILTER_SIZE_H_BPERC){
             clientMsg = getPart_H_B(f, msg, numBytes);
          } else if (sizeFilter == Def.FILTER_SIZE_H_B_A){
             clientMsg = getPart_H_B_A(f, msg);
          } else if (sizeFilter == Def.FILTER_SIZE_H_B_APERC){
             clientMsg = getPart_H_B_A(f, msg, numBytes);
          }
       }
       return clientMsg;
    }


}
