/**
 * Copyright (C) 2005-2006 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.email.pdi.mail;


import com.funambol.foundation.pdi.common.Property;

/**
 * An object representing a email.
 * This class can be used to store and retrive all
 * information about email.
 *
 * @version $Id: Email.java,v 1.2 2006/02/06 09:03:01 gbmiglia Exp $
 */
public class Email {

    private Property  uid;       // String  --> will be used String type
    private Property  read;      // boolean --> will be used String type
    private Property  forwarded; // boolean --> will be used String type
    private Property  replied;   // boolean --> will be used String type
    private Property  received;  // date    --> will be used String type
    private Property  created;   // date    --> will be used String type
    private Property  modified;  // date    --> will be used String type
    private Property  deleted;   // boolean --> will be used String type
    private Property  flagged;   // boolean --> will be used String type
    private EmailItem emailItem;
    // @todo Ext


    /**
     * Creates an empty mail
     */
    public Email() {
        uid       = new Property();
        read      = new Property();
        forwarded = new Property();
        replied   = new Property();
        received  = new Property();
        created   = new Property();
        modified  = new Property();
        deleted   = new Property();
        flagged   = new Property();
        emailItem = new EmailItem();
    }

    /**
     * Returns the uid of this mail
     *
     * @return the uid of this mail
     */
    public Property getUID() {
        return uid;
    }

    /**
     * set the uid of this mail
     *
     * @param uid Property
     */
    public void setUID(Property uid) {
        this.uid = uid;
    }

    /**
     * Returns the read of this mail
     *
     * @return the read of this mail
     */
    public Property getRead() {
        return read;
    }

    /**
     * Returns the forwarded of this mail
     *
     * @return the forwarded of this mail
     */
    public Property getForwarded() {
        return forwarded;
    }

    /**
     * Returns the replied of this mail
     *
     * @return the replied of this mail
     */
    public Property getReplied() {
        return replied;
    }

    /**
     * Returns the received of this mail
     *
     * @return the received of this mail
     */
    public Property getReceived() {
        return received;
    }

    /**
     * Returns the created of this mail
     *
     * @return the created of this mail
     */
    public Property getCreated() {
        return created;
    }

    /**
     * Returns the modified of this mail
     *
     * @return the modified of this mail
     */
    public Property getModified() {
        return modified;
    }

    /**
     * Returns the deleted of this mail
     *
     * @return the deleted of this mail
     */
    public Property getDeleted() {
        return deleted;
    }

    /**
     * Returns the flagged of this mail
     *
     * @return the flagged of this mail
     */
    public Property getFlagged() {
        return flagged;
    }

    /**
     * Returns the mailItem of this mail
     *
     * @return the mailItem of this mail
     */
    public EmailItem getEmailItem() {
        return emailItem;
    }

    /**
     *
     * @return String
     */
    public String toString() {
        StringBuffer sb = new StringBuffer("Email:");
        sb.append("\nUID       : " + this.getUID().getPropertyValue());
        sb.append("\nRead      : " + this.getRead().getPropertyValue());
        sb.append("\nForwarded : " + this.getForwarded().getPropertyValue());
        sb.append("\nReplied   : " + this.getReplied().getPropertyValue());
        sb.append("\nReceived  : " + this.getReceived().getPropertyValue());
        sb.append("\nCreated   : " + this.getCreated().getPropertyValue());
        sb.append("\nModified  : " + this.getModified().getPropertyValue());
        sb.append("\nDeleted   : " + this.getDeleted().getPropertyValue());
        sb.append("\nDeleted   : " + this.getFlagged().getPropertyValue());
        sb.append("\nEmailItem : " + this.emailItem.toString());
        return sb.toString();
    }
}
