/**
 * Copyright (C) 2005-2006 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.email.transport  ;

import java.util.logging.Level;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.Properties;
import java.util.Locale;
import java.security.Security;
import javax.mail.Session;
import com.sun.net.ssl.internal.ssl.Provider;
import com.sun.mail.imap.IMAPFolder;

import com.funambol.email.exception.EmailAccessException;
import com.funambol.email.util.Def;


/**
 *
 * <p>
 * Create a connection with a Mail Server
 * using the IMAP protocol
 * </p>
 *
 * @version $Id: ImapMailServerWrapper.java,v 1.2 2006/02/06 09:03:37 gbmiglia Exp $
 */
public class ImapMailServerWrapper
        extends CommonMailServerWrapper
        implements IMailServerWrapper
{

    // -------------------------------------------------------------- Properties

    /**     */
    private IMAPFolder folderInbox ;
    /**     */
    private IMAPFolder folderOutbox ;
    /**     */
    private IMAPFolder folderDrafts ;
    /**     */
    private IMAPFolder folderSent ;
    /**     */
    private IMAPFolder folderTrash ;
    /**     */
    private String labelInbox;
    /**     */
    private String labelOutbox;
    /**     */
    private String labelDrafts;
    /**     */
    private String labelSentItems;
    /**     */
    private String labelDeletedItems;

    // ---------------------------------------------------------- PUBLIC METHODS


    /**
     * opens the connection with the mail server
     * outgoingServer: 192.168.0.1 (SMTP)
     * incomingServer: 192.168.0.1 (IMAP)
     * incomingPort: 143 / 993 IMAP def.
     * language: en, it, fr, ...
     * user:
     * password:
     * isSSL:  true/false
     *
     * @param outgoingServer String
     * @param incomingServer String
     * @param incomingPort String
     * @param language String
     * @param user String
     * @param password String
     * @param isSSL boolean
     * @throws EmailAccessException
     */
    public void openConnection(String  outgoingServer,
                               String  incomingServer,
                               String  incomingPort,
                               String  language,
                               String  user,
                               String  password,
                               boolean isSSL)
      throws EmailAccessException {

      if (log.isLoggable(Level.FINEST)) {
          log.finest("\nCONNECTION PARAMs"+
                     "\nin           " + outgoingServer+
                     "\nout          " + incomingServer+
                     "\nport         " + incomingPort+
                     "\nprotocol     " + this.getClass().getName()+
                     "\nuser         " + user+
                     "\nSSL          " + isSSL);
      }

      // containe for properties
      Properties props = new Properties();

      // provider for a SSL connection
      Provider provider = null;


      if (isSSL) {
        this.setConnectionSSL(props,
                              outgoingServer,
                              incomingServer,
                              incomingPort,
                              provider);
      } else {
        this.setConnectionNormal(props,
                                 outgoingServer,
                                 incomingServer,
                                 incomingPort);
      }

      try{
         this.session = Session.getInstance(props);
         //this.session.setDebug(true);
         this.store = session.getStore();
         this.store.connect(incomingServer, user, password);
      } catch (Exception e) {
         log.severe("Error Opening Connection Store: " + e);
         log.throwing(this.getClass().getName(), "openConnection", e);
	     throw new EmailAccessException("Error Opening Connection Store (" + e.getMessage() + ")");
      }

      // set locale
      try {

         this.locale = new Locale(language);

         // set the name of the folder
         ResourceBundle messages = ResourceBundle.getBundle(RESOURCE_NAME, this.locale);
         this.labelInbox        = messages.getString(Def.LABEL_INBOX);
         this.labelOutbox       = messages.getString(Def.LABEL_OUTBOX);
         this.labelDrafts       = messages.getString(Def.LABEL_DRAFTS);
         this.labelSentItems    = messages.getString(Def.LABEL_SENT);
         this.labelDeletedItems = messages.getString(Def.LABEL_TRASH);

         // set the folder
         // this folder is always in the mail store
         this.folderInbox  = (IMAPFolder)store.getFolder(this.labelInbox);
         // about this folders we have to check
         this.folderDrafts = (IMAPFolder)store.getFolder(this.labelDrafts);
         this.folderSent   = (IMAPFolder)store.getFolder(this.labelSentItems);
         this.folderOutbox = (IMAPFolder)store.getFolder(this.labelOutbox);
         this.folderTrash  = (IMAPFolder)store.getFolder(this.labelDeletedItems);

      } catch (Exception e) {
         log.severe("Error getting Folders: " + e);
         log.throwing(this.getClass().getName(), "openConnection", e);
	     throw new EmailAccessException("Error getting Folders (" + e.getMessage() + ")");
      }

      if (log.isLoggable(Level.FINEST)) {
          log.finest("IMAP end opening Connection");
      }

    }


    /**
     * closes the connection with the mail server
     *
     * @throws EmailAccessException
     */
    public void releaseConnection() throws EmailAccessException {
        try {
            this.store.close();
            this.store = null;
            this.session = null;
        } catch (Exception e) {
     	    throw new EmailAccessException("Error Closing Store " + e.toString());
        }

    }

    /**
     *
     *
     * @return String
     */
    public String getInboxFolderName(){
       return this.labelInbox;
    }

    /**
     *
     *
     * @return String
     */
    public String getOutboxFolderName(){
      return this.labelOutbox;
    }

    /**
     *
     *
     * @return String
     */
    public String getDraftsFolderName(){
       return this.labelDrafts;
    }

    /**
     *
     *
     * @return String
     */
    public String getSentItemsFolderName(){
       return this.labelSentItems;
    }

    /**
     *
     *
     * @return String
     */
    public String getDeletedItemsFolderName(){
        return this.labelDeletedItems;
    }


    /**
     *
     *
     * @return Inbox Folder IMAPFolder
     */
    public IMAPFolder getInboxFolder(){
       return this.folderInbox;
    }

    /**
     *
     *
     * @return Outbox Folder IMAPFolder
     */
    public IMAPFolder getOutboxFolder(){
       return this.folderOutbox;
    }

    /**
     *
     *
     * @return Drafts Folder IMAPFolder
     */
    public IMAPFolder getDraftsFolder(){
       return this.folderDrafts;
    }

    /**
     *
     *
     * @return SentItems Folder IMAPFolder
     */
    public IMAPFolder getSentFolder(){
       return this.folderSent;
    }

    /**
     *
     *
     * @return DeletedItems Folder IMAPFolder
     */
    public IMAPFolder getTrashFolder(){
       return this.folderTrash;
    }

    //---------------------------------------------------------- PRIVATE METHODS

   /**
    *
    * @param props Properties
    * @param outgoingServer String
    * @param incomingServer String
    * @param incomingPort String
    * @param provider Provider
    */
   protected void setConnectionSSL(Properties props,
                                 String outgoingServer,
                                 String incomingServer,
                                 String incomingPort,
                                 Provider provider){

        if (log.isLoggable(Level.FINEST)) {
           log.finest("IMAP set an SSL connection");
        }
        //
        props.setProperty("mail.smtp.host" , outgoingServer);
        props.setProperty("mail.transport.protocol" , "smtp");

        props.setProperty("mail.imap.host" , incomingServer);
        props.setProperty("mail.imap.port" , incomingPort);

        props.setProperty("mail.store.protocol", "imap");

        //
        System.setProperty("java.protocol.handler.pkgs",
                                  "com.sun.net.ssl.internal.www.protocol");
        System.setProperty("javax.net.ssl.trustStore",
                                  "clientExchange.keystore");

        provider = new Provider();
	    Security.addProvider(provider);

        // set SSL parameters
    	props.setProperty("mail.imap.socketFactory.class",
                                 "javax.net.ssl.SSLSocketFactory");
    	props.setProperty("mail.imap.socketFactory.fallback",
                                 "false");
    	props.setProperty("mail.imap.socketFactory.port",
                                 incomingPort);

    }

    /**
     *
     *
     * @param props Properties
     * @param outgoingServer String
     * @param incomingServer String
     * @param incomingPort String
     */
    protected void setConnectionNormal(Properties props,
                                    String outgoingServer,
                                    String incomingServer,
                                    String incomingPort){

        if (log.isLoggable(Level.FINEST)) {
           log.finest("IMAP set a normal connection");
        }
        //
        props.setProperty("mail.smtp.host", outgoingServer);
        props.setProperty("mail.transport.protocol", "smtp");

        props.setProperty("mail.imap.host"    , incomingServer);
        props.setProperty("mail.imap.port"    , incomingPort);

        props.setProperty("mail.store.protocol", "imap");
    }

}
