/**
 * Copyright (C) 2005-2006 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.email.transport;

import java.util.logging.Level;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.Properties;
import java.util.Locale;
import java.security.Security;
import javax.mail.Session;
import com.sun.net.ssl.internal.ssl.Provider;
import com.sun.mail.pop3.POP3Folder;

import com.funambol.email.exception.EmailAccessException;
import com.funambol.email.util.Def;


/**
 *
 * <p>Create a connection with a Mail Server </p>
 *
 * @version $Id: PopMailServerWrapper.java,v 1.2 2006/02/06 09:03:37 gbmiglia Exp $
 */
public class PopMailServerWrapper
        extends CommonMailServerWrapper
        implements IMailServerWrapper
{

    // -------------------------------------------------------------- Properties

    /**     */
    private POP3Folder folderInbox ;
    /**     */
    private String labelInbox;

    // ---------------------------------------------------------- PUBLIC METHODS


    /**
     * opens the connection with the mail server
     * outgoingServer: 192.168.0.1 (SMTP)
     * incomingServer: 192.168.0.1 (POP3)
     * incomingPort: 110 / 995 POP3 def.
     * language: en, it, fr, ...
     * user
     * password
     * isSSL: true/false
     *
     *
     * @param outgoingServer String
     * @param incomingServer String
     * @param incomingPort String
     * @param language String
     * @param user String
     * @param password String
     * @param isSSL boolean
     * @throws EmailAccessException
     */
    public void openConnection(String  outgoingServer, // 192.168.0.1 (SMTP)
                               String  incomingServer, // 192.168.0.1 (POP3)
                               String  incomingPort,   // 110 / 995 POP3 def.
                               String  language,       // en, it, fr, ...
                               String  user,           //
                               String  password,       //
                               boolean isSSL)          // true/false
      throws EmailAccessException {


      if (log.isLoggable(Level.FINEST)) {
          log.finest("\nCONNECTION PARAMs"+
                     "\nin           " + outgoingServer+
                     "\nout          " + incomingServer+
                     "\nport         " + incomingPort+
                     "\nprotocol     " + this.getClass().getName()+
                     "\nuser         " + user+
                     "\nSSL          " + isSSL);
      }


      // containe for properties
      Properties props = new Properties();

      // provider for a SSL connection
      Provider provider = null;

      if (isSSL) {
        this.setConnectionSSL(props,
                              outgoingServer,
                              incomingServer,
                              incomingPort,
                              provider);
      } else {
        this.setConnectionNormal(props,
                                 outgoingServer,
                                 incomingServer,
                                 incomingPort);
      }

      if (log.isLoggable(Level.FINEST)) {
          log.finest("POP property: " + props.getProperty("mail.store.protocol"));
      }

      try{
         this.session = Session.getInstance(props);
         //this.session.setDebug(true);
  	     this.store = session.getStore();
         this.store.connect(incomingServer, user, password);
      } catch (Exception e) {
         log.severe("Error Opening Connection Store: " + e);
         log.throwing(this.getClass().getName(), "openConnection", e);
	     throw new EmailAccessException("Error Opening Connection Store (" + e.getMessage() + ")");
      }

      // set locale and folder name
      try {
         this.locale = new Locale(language);
         ResourceBundle messages = ResourceBundle.getBundle(RESOURCE_NAME, this.locale);
         this.labelInbox        = messages.getString(Def.LABEL_INBOX);
         // set the folder
         this.folderInbox  = (POP3Folder)store.getFolder(this.labelInbox);
      } catch (Exception e) {
         log.severe("Error getting Folders: " + e);
         log.throwing(this.getClass().getName(), "openConnection", e);
	     throw new EmailAccessException("Error getting Folders (" + e.getMessage() + ")");
      }

      if (log.isLoggable(Level.FINEST)) {
          log.finest("POP end opening Connection");
      }


    }



    /**
     * closes the connection with the mail server
     *
     * @throws EmailAccessException
     */
    public void releaseConnection() throws EmailAccessException {
        try {
            this.store.close();
            this.store = null;
            this.session = null;
        } catch (Exception e) {
     	    throw new EmailAccessException("Error Closing Store " + e.toString());
        }

    }

    /**
     *
     *
     * @return Inbox Folder POP3Folder
     */
    public POP3Folder getInboxFolder(){
       return this.folderInbox;
    }


    /**
     *
     *
     * @return String
     */
    public String getInboxFolderName(){
       return this.labelInbox;
    }

    //---------------------------------------------------------- PRIVATE METHODS

   /**
    *
    *
    * @param props Properties
    * @param outgoingServer String
    * @param incomingServer String
    * @param incomingPort String
    * @param provider Provider
    */
   protected void setConnectionSSL(Properties props,
                                 String outgoingServer,
                                 String incomingServer,
                                 String incomingPort,
                                 Provider provider){

        if (log.isLoggable(Level.FINEST)) {
           log.finest("POP3 set an SSL connection");
        }
        //
        props.setProperty("mail.smtp.host" , outgoingServer);
        props.setProperty("mail.transport.protocol" , "smtp");

        props.setProperty("mail.pop3.host" , incomingServer);
        props.setProperty("mail.pop3.port" , incomingPort);

        props.setProperty("mail.store.protocol", "pop3");

        //
        System.setProperty("java.protocol.handler.pkgs",
                                      "com.sun.net.ssl.internal.www.protocol");
        System.setProperty("javax.net.ssl.trustStore",
                                      "clientExchange.keystore");

        provider = new Provider();
	    Security.addProvider(provider);

        // set SSL parameters
    	props.setProperty("mail.pop3.socketFactory.class",
                                     "javax.net.ssl.SSLSocketFactory");
    	props.setProperty("mail.pop3.socketFactory.fallback",
                                     "false");
    	props.setProperty("mail.pop3.socketFactory.port",
                                     incomingPort);

    }

    /**
     *
     *
     * @param props Properties
     * @param outgoingServer String
     * @param incomingServer String
     * @param incomingPort String
     */
    protected void setConnectionNormal(Properties props,
                                    String outgoingServer,
                                    String incomingServer,
                                    String incomingPort){
        if (log.isLoggable(Level.FINEST)) {
           log.finest("POP3 set a normal connection");
        }
        //
        props.setProperty("mail.smtp.host", outgoingServer);
        props.setProperty("mail.transport.protocol", "smtp");

        props.setProperty("mail.pop3.host"    , incomingServer);
        props.setProperty("mail.pop3.port"    , incomingPort);

        props.setProperty("mail.store.protocol", "pop3");
    }


}
