/**
 * Copyright (C) 2005-2006 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.email.util;


import java.text.SimpleDateFormat;
import java.text.DateFormat;
import java.sql.Timestamp;
import java.util.Enumeration;
import java.util.Date;
import java.util.TimeZone;
import java.util.Locale;
import java.util.logging.Logger;
import java.util.zip.CRC32;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import javax.mail.Message;
import javax.mail.Header;
import javax.mail.Session;
import javax.mail.internet.MimeMessage;
import javax.mail.MessagingException;
import javax.mail.Flags;
import javax.activation.DataHandler;
import com.sun.mail.imap.IMAPFolder;
import com.sun.mail.pop3.POP3Folder;

import com.funambol.framework.tools.Base64;
import com.funambol.framework.engine.SyncItem;
import com.funambol.foundation.common.TimeZonesTools;

import com.funambol.email.exception.EntityException;
import com.funambol.email.transport.ImapMailServerWrapper;
import com.funambol.email.transport.PopMailServerWrapper;
import com.funambol.email.items.model.FlagProperties;
import com.funambol.email.cache.CrcSyncItemInfo;



/**
 *
 * <p>Utility class</p>
 *
 * @version $Id: Utility.java,v 1.2 2006/02/06 09:03:47 gbmiglia Exp $
 */

public class Utility {

  // -------------------------------------------------------------- Private data



  // ------------------------------------------------------------ Public Methods

  /**
   * credantials format:  "username:passowrd"
   *
   * @param username String
   * @param credentials String
   * @return password String
   */
  public static String getPassword(String username, String credentials){
      String up = new String (Base64.decode(credentials));
      int l = username.length();
      String password = up.substring(l+1,up.length());
      return password;
  }

  /**
   * get content from a SyncItem
   *
   *
   * @param syncItem SyncItem
   * @param isEncode boolean
   * @return String
   * @throws EntityException
   */
  public static String getContentFromSyncItem(SyncItem syncItem,
                                              boolean isEncode)
  throws EntityException {
     String content = null;

     // get content of the item
     try {

        byte[] itemContent = (byte[])syncItem.getContent();

        if (itemContent == null) {
            itemContent = new byte[0];
        }
        content = new String (itemContent);
        if (isEncode &&
            content != null &&
            content.length() > 0) {
            content = new String (Base64.decode(content));
        }
     } catch (Exception e){
        throw new EntityException("Error getting the item "+syncItem,e);
     }

     return content;
  }

  /**
   *
   *
   * @param message Message
   * @return String
   */
  public static String getMessageID(Message message){
     try {
         String[] values = message.getHeader(Def.HEADER_LABEL_MESSAGE_ID);
         if ((values != null) && (values.length > 0)) {
             return values[0];
         }
     } catch (MessagingException exception) {
         // do nothing; we just want to return null
     }
     return null;
  }


  /**
   *
   *
   * @param message Message
   * @return String
   */
  public static String getSyncLabel(Message message){
     try {
         String[] values = message.getHeader(Def.HEADER_LABEL_SYNCMESSAGE);
         if ((values != null) && (values.length > 0)) {
             return values[0];
         }
     } catch (MessagingException exception) {
         // do nothing; we just want to return null
     }
     return null;
  }

  /**
   * Date
   * pat_1: Exchange 2003 (eng): wed, 6 Oct 2005 09:46:47 +0200
   * pat_2: Exchange 2000 (eng): 6 Oct 2005 09:46:47 +0200
   * pat_3: BB client 7230     : Wed Dec 28 20:05:53 GMT 2005
   * pat_4: BB client 7290     : Wed Feb 01 10:49:21 Europe/Amsterdam 2006
   *
   * Returns an origination date in UTC format
   *
   * @param message Message
   * @param loc Locale
   * @return String
   */
  public static Date getOriginationDate(Message message, Locale loc){

      /**
       * the date inside MS Exchange are in US/UK Locale
       */
      loc = Locale.US;

      DateFormat pat_1 = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss Z", loc);
      DateFormat pat_2 = new SimpleDateFormat("dd MMM yyyy HH:mm:ss Z", loc);
      DateFormat pat_3 = new SimpleDateFormat("EEE MMM dd HH:mm:ss Z yyyy", loc);
      DateFormat pat_4 = new SimpleDateFormat("EEE MMM dd HH:mm:ss z yyyy", loc);

      String out      = null;
      Date origDate   = null;
      String[] values = null;

      try {
        values = message.getHeader(Def.HEADER_LABEL_DATE);
      } catch (MessagingException me){
        me.printStackTrace();
      }

      if ((values != null) && (values.length > 0)) {

          String s = values[0];

          //System.out.println("TO DELETE partenza 1:" + s);

          s = bb7290toRFC822(s);

          //System.out.println("TO DELETE partenza 2:" + s);

		  try{
			  origDate = pat_1.parse(s);
		  } catch(Exception e1){
			  try{
				origDate = pat_2.parse(s);
			  }	catch(Exception e2){
    			  try{
	    			origDate = pat_3.parse(s);
		    	  }	catch(Exception e3){
        			  try{
    	    			origDate = pat_4.parse(s);
    		    	  }	catch(Exception e4){
    			        e4.printStackTrace();
        			  }
    			  }
			  }
		  }

          //System.out.println("TO DELETE origDate:" + origDate);

      }


      return origDate;
  }

  /**
   * Date
   * pat_1: Exchange 2003 (eng): wed, 6 Oct 2005 09:46:47 +0200
   * pat_2: Exchange 2000 (eng): 6 Oct 2005 09:46:47 +0200
   * pat_3: BB client 7230     : Wed Dec 28 20:05:53 GMT 2005
   * pat_4: BB client 7290     : Wed Feb 01 10:49:21 Europe/Amsterdam 2006
   *
   * Returns an origination date in UTC format
   *
   *
   * @param origDate Date
   * @param loc Locale
   * @return String
   */
  public static String DateToUTC(Date origDate, Locale loc){

      String out = null;

      //System.out.println("TO DELETE origDate:" + origDate);

      // convert in UTC format
      //DateFormat utcf = new SimpleDateFormat("yyyyMMdd'T'HHmmss'Z'", loc);
      DateFormat utcf = new SimpleDateFormat("yyyyMMdd'T'HHmmss'Z'");
      utcf.setTimeZone(TimeZone.getTimeZone("UTC"));
      if (origDate != null){
          out = utcf.format(origDate);
      }

      //System.out.println("TO DELETE utc" + out);

      return out;
  }

  /**
   * Date : Thu Oct 06 09:46:47 CEST 2005
   * Returns a received date in UTC format
   *
   *
   * @param dateold String
   * @return String
   */
  public static String bb7290toRFC822(String dateold){

     String rfc822 = null;

     int index     = 0;
     String label  = "";
     String offset = "";

     String[] tmzsDescr = TimeZonesTools.getAvailablesTimeZonesDescr();

     for (int i=0; i<tmzsDescr.length; i++){

        index = tmzsDescr[i].indexOf("(GMT");

        if (index != -1){

           label  = tmzsDescr[i].substring(0,index-1);
           offset = tmzsDescr[i].substring(index+1,tmzsDescr[i].length()-1);

           index  = dateold.indexOf(label);

           if (index != -1){
              // is a bb_7290
              rfc822 = dateold.replaceAll(label,offset);
           } else {
              //
              rfc822 = dateold;
           }

        } else {
           // is not a bb_7290
           rfc822 = dateold;
        }

     }

     return  rfc822;

  }



  /**
   *   Date : Thu Oct 06 09:46:47 CEST 2005
   *   Returns a received date in UTC format
   *
   * @param message Message
   * @param loc Locale
   * @return String
   */
  public static String getReceivedDate(Message message, Locale loc){

      String out = null;

      try {
         Date d = message.getReceivedDate();

         // convert in UTC format
         DateFormat utcf = new SimpleDateFormat("yyyyMMdd'T'HHmmss'Z'", loc);
         utcf.setTimeZone(TimeZone.getTimeZone("UTC"));

         out = utcf.format(d);

      } catch (Exception e) {
         // do nothing; we just want to return null
      }
      return out;
  }


  /**
   * sets the folder target id
   *
   *
   * @param key Sync Item key Value String
   * @return String
   */
  public static String getFID(String key){
     int endFolderId = key.indexOf(Def.SEPARATOR);
     String folderID = key.substring(0,endFolderId);
     return folderID;
  }


  /**
   *
   *
   * @param key String
   * @return String
   */
  public static String getFMID(String key){
     int endFolderId = key.indexOf(Def.SEPARATOR);
     String FMID = key.substring((endFolderId + 1),key.length());
     return FMID;
  }


  /**
   *
   *
   * @param f IMAPFolder
   * @param imsw ImapMailServerWrapper
   * @return String
   */
  public static String getFID(IMAPFolder f, ImapMailServerWrapper imsw){

     String folderID = null;

     if        ( f.getName().equalsIgnoreCase(imsw.getInboxFolderName())){
        folderID = Def.FOLDER_INBOX_ID;
     } else if ( f.getName().equalsIgnoreCase(imsw.getOutboxFolderName())){
        folderID = Def.FOLDER_OUTBOX_ID;
     } else if ( f.getName().equalsIgnoreCase(imsw.getDraftsFolderName())){
        folderID = Def.FOLDER_DRAFTS_ID;
     } else if ( f.getName().equalsIgnoreCase(imsw.getSentItemsFolderName())){
        folderID = Def.FOLDER_SENT_ID;
     } else if ( f.getName().equalsIgnoreCase(imsw.getDeletedItemsFolderName())){
        folderID = Def.FOLDER_TRASH_ID;
     }

     return folderID;
  }

  /**
   *
   *
   * @param f POP3Folder
   * @param pmsw PopMailServerWrapper
   * @return String
   */
  public static String getFID(POP3Folder f, PopMailServerWrapper pmsw){

     String folderID = null;

     if ( f.getName().equalsIgnoreCase(pmsw.getInboxFolderName())){
        folderID = Def.FOLDER_INBOX_ID;
     }

     return folderID;
  }




  /**
   *
   *
   * @param FID String
   * @param imsw ImapMailServerWrapper
   * @return IMAPFolder
   */
  public static IMAPFolder getFolderIMAP(String FID,
                                         ImapMailServerWrapper imsw){

      IMAPFolder f = null;

      if        ( Def.FOLDER_INBOX_ID.equalsIgnoreCase(FID)){
           f = imsw.getInboxFolder();
      } else if (Def.FOLDER_OUTBOX_ID.equalsIgnoreCase(FID)){
           f = imsw.getOutboxFolder();
      } else if (Def.FOLDER_DRAFTS_ID.equalsIgnoreCase(FID)){
           f = imsw.getDraftsFolder();
      } else if (Def.FOLDER_SENT_ID.equalsIgnoreCase(FID)){
           f = imsw.getSentFolder();
      } else if (Def.FOLDER_TRASH_ID.equalsIgnoreCase(FID)){
           f = imsw.getTrashFolder();
      }

      return f;
  }


  /**
   *
   *
   * @param FID String
   * @param pmsw PopMailServerWrapper
   * @return POP3Folder
   */
  public static POP3Folder getFolderPOP3(String FID,
                                         PopMailServerWrapper pmsw){

      POP3Folder f = null;

      if ( Def.FOLDER_INBOX_ID.equalsIgnoreCase(FID)){
         f = pmsw.getInboxFolder();
      }

      return f;
  }

  /**
   *  convert the UTC string in a Timestamp
   *
   * @param utc String
   * @return Timestamp
   */
  public static Timestamp UtcToTimestamp(String utc){

      Timestamp t = null;
      if (utc != null){
         try {
            DateFormat utcFormatter = new SimpleDateFormat("yyyyMMdd'T'HHmmss'Z'");
            Date d = utcFormatter.parse(utc);
            t = new Timestamp(d.getTime());
         } catch (Exception e) {
            e.printStackTrace();
         }
      }
      return t;
  }


  /**
   *  convert the UTC string in a Date
   *
   * @param utc String
   * @return Date
   */
  public static Date UtcToDate(String utc){

      Date d = null;
      if (utc != null){
         try {
            DateFormat utcFormatter = new SimpleDateFormat("yyyyMMdd'T'HHmmss'Z'");
            d = utcFormatter.parse(utc);
         } catch (Exception e) {
            e.printStackTrace();
         }
      }
      return d;
  }


  /**
   *  convert String (RFC2822) into Message
   *
   *
   * @param session Session
   * @param s String
   * @return Message
   */
  public static Message StreamToMessage(Session session, String s) {

      Message m = null;
      if (s != null){
         try {
            byte[] b = s.getBytes();
            InputStream is = new ByteArrayInputStream(b);
            m = new MimeMessage(session, is);
         } catch (Exception e) {
            e.printStackTrace();
         }
      }
      return m;

  }


  /**
   *  convert Message in a String (RFC2822)
   *
   *
   * @param msg Message
   * @return String
   */
  public static String messageToString(Message msg) {

      String s = null;

      if (msg != null){
         try {
            boolean isSeen = msg.isSet(Flags.Flag.SEEN);

            OutputStream os = new ByteArrayOutputStream();
            msg.writeTo(os);
            s = os.toString();

            msg.setFlag(Flags.Flag.SEEN, isSeen);

         } catch (Exception e) {
            e.printStackTrace();
         }
      }

      return s;

  }

  /**
   *
   * merge arrays
   *
   * @param i CrcSyncItemInfo[]
   * @param o CrcSyncItemInfo[]
   * @param d CrcSyncItemInfo[]
   * @param s CrcSyncItemInfo[]
   * @param t CrcSyncItemInfo[]
   * @return CrcSyncItemInfo[]
   */
  public static String[] mergeArray(String[] i,
                                    String[] o,
                                    String[] d,
                                    String[] s,
                                    String[] t){

        String[] ids = null;

        int length  = 0;
        int ilength = 0;
        int olength = 0;
        int dlength = 0;
        int slength = 0;
        int tlength = 0;

        if (i!=null){
           ilength = i.length;
           length = ilength;
        }
        if (o!=null){
           olength = o.length;
           length = length + olength;
        }
        if (d!=null){
           dlength = d.length;
           length = length + dlength;
        }
        if (s!=null){
           slength = s.length;
           length = length + slength;
        }
        if (t!=null){
           tlength = t.length;
           length = length + tlength;
        }

        ids = new String[length];

        int yi = 0;
        int yo = 0;
        int yd = 0;
        int ys = 0;
        int yt = 0;

        for (yi = 0; yi < ilength; yi++){
            ids[yi] = i[yi];
        }
        for (yo = 0; yo < olength; yo++){
            ids[yi+yo] = o[yo];
        }
        for (yd = 0; yd < dlength; yd++){
            ids[yi+yo+yd] = d[yd];
        }
        for (ys = 0; ys < slength; ys++){
            ids[yi+yo+yd+ys] = s[ys];
        }
        for (yt = 0; yt < tlength; yt++){
            ids[yi+yo+yd+ys+yt] = t[yt];
        }

        return ids;
  }

  /**
   *
   * merge arrays
   *
   * @param i CrcSyncItemInfo[]
   * @param o CrcSyncItemInfo[]
   * @param d CrcSyncItemInfo[]
   * @param s CrcSyncItemInfo[]
   * @param t CrcSyncItemInfo[]
   * @return CrcSyncItemInfo[]
   */
  public static CrcSyncItemInfo[] mergeArray(CrcSyncItemInfo[] i,
                                          CrcSyncItemInfo[] o,
                                          CrcSyncItemInfo[] d,
                                          CrcSyncItemInfo[] s,
                                          CrcSyncItemInfo[] t){
        CrcSyncItemInfo[] syncItemInfo = null;

        int length  = 0;
        int ilength = 0;
        int olength = 0;
        int dlength = 0;
        int slength = 0;
        int tlength = 0;

        if (i!=null){
           ilength = i.length;
           length = ilength;
        }
        if (o!=null){
           olength = o.length;
           length = length + olength;
        }
        if (d!=null){
           dlength = d.length;
           length = length + dlength;
        }
        if (s!=null){
           slength = s.length;
           length = length + slength;
        }
        if (t!=null){
           tlength = t.length;
           length = length + tlength;
        }

        syncItemInfo = new CrcSyncItemInfo[length];

        int yi = 0;
        int yo = 0;
        int yd = 0;
        int ys = 0;
        int yt = 0;

        for (yi = 0; yi < ilength; yi++){
            syncItemInfo[yi] = i[yi];
        }
        for (yo = 0; yo < olength; yo++){
            syncItemInfo[yi+yo] = o[yo];
        }
        for (yd = 0; yd < dlength; yd++){
            syncItemInfo[yi+yo+yd] = d[yd];
        }
        for (ys = 0; ys < slength; ys++){
            syncItemInfo[yi+yo+yd+ys] = s[ys];
        }
        for (yt = 0; yt < tlength; yt++){
            syncItemInfo[yi+yo+yd+ys+yt] = t[yt];
        }

        return syncItemInfo;
  }


  /**
   *
   * @param msg Message
   * @return long
   */
  public static long setLastCrc(Message msg) {

      // Compute CRC-32 checksum
      CRC32 checksumEngine = new CRC32();
      long checksumFlag    = 0;
      long checksumMessage = 0;

      // set CRC for the Flag properties
      try {

         boolean isSeen     = msg.isSet(Flags.Flag.SEEN);
         boolean isAnswered = msg.isSet(Flags.Flag.ANSWERED);
         boolean isDeleted  = msg.isSet(Flags.Flag.DELETED);
         boolean isDraft    = msg.isSet(Flags.Flag.DRAFT);
         boolean isRecent   = msg.isSet(Flags.Flag.RECENT);
         boolean isFlagged  = msg.isSet(Flags.Flag.FLAGGED);

         FlagProperties fg = new FlagProperties(isSeen,
                                                isAnswered,
                                                isDeleted,
                                                isDraft,
                                                isRecent,
                                                isFlagged);

         String fgS = fg.toString();
         byte[] bytesFlags = fgS.getBytes();
         checksumEngine.update(bytesFlags, 0, bytesFlags.length);
         checksumFlag = checksumEngine.getValue();
         checksumEngine.reset();


         String msgS = messageToString(msg);
         byte[] bytesMsg = msgS.getBytes();
         checksumEngine.update(bytesMsg, 0, bytesMsg.length);
         checksumMessage = checksumEngine.getValue();
         checksumEngine.reset();


      } catch (MessagingException me){
         me.printStackTrace();
      }

      return  checksumFlag + checksumMessage;

  }

    /**
     * set a new header but keep the old Message-ID
     *
     * @param newMsg Message
     * @param oldMsg Message
     * @return String
     * @throws EntityException
     */
    public static String setNewHeader(Message newMsg, Message oldMsg)
      throws EntityException {


      String oldMessageID = null;

      try {

         Header h     = null;
         String name  = null;
         String value = null;

         Enumeration en = oldMsg.getAllHeaders();
         while (en.hasMoreElements()) {
              h = (Header) en.nextElement();
              name  = h.getName();
              value = h.getValue();
              if (name.equalsIgnoreCase(Def.HEADER_LABEL_MESSAGE_ID)){
                 oldMessageID = value;
              }
              newMsg.setHeader(name, value);
         }

      } catch (MessagingException me){
          throw new EntityException(me);
      }

      return oldMessageID;

    }


   /**
    * TMP: only for debug purpose
    *
    * @param msg Message
    */
   public static void printMessage(Message msg) {

     try {
       // header
       Enumeration en = msg.getAllHeaders();
       while ( en.hasMoreElements() ){
          Header h = (Header)en.nextElement();
          System.out.println(" " + h.getName() + " : " + h.getValue());
       }

       // content
       String contentType = msg.getContentType();
       if (contentType != null){
          if (contentType.startsWith("text/html") ){
          }
          if (contentType.startsWith("text/html") ){
          }
       } else {
          DataHandler dh = (DataHandler)msg.getDataHandler();
          System.out.println(" " );
          System.out.println(" " + dh.getContent());
       }


       System.out.println(" --------------------------------------------");
       System.out.println(" id       : " + getMessageID(msg));
       System.out.println(" seen     : " + msg.isSet(Flags.Flag.SEEN));
       // System.out.println(" from     : " + msg.getFrom()[0]);
       System.out.println(" subj     : " + msg.getSubject());
       System.out.println(" cont     : " + msg.getContent());
       System.out.println(" type     : " + msg.getContentType());
       System.out.println(" received : " + msg.getReceivedDate());
       System.out.println(" sent     : " + msg.getSentDate());
       System.out.println(" ---------------------------------------------");
     }catch (Exception e){
       e.printStackTrace();
     }

   }

  //------------------------------------------------------------ PRIVATE METHODS

}
